# /*##########################################################################
# Copyright (C) 2022-2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
import contextlib
from io import BytesIO

import h5py
import numpy
import pytest

import silx.io
from silx.io import commonh5
from silx.io._sliceh5 import DatasetSlice, _combine_indices


@contextlib.contextmanager
def h5py_file(filename, mode):
    with BytesIO() as buffer:
        with h5py.File(buffer, mode) as h5file:
            yield h5file


@pytest.fixture(params=[commonh5.File, h5py_file])
def temp_h5file(request):
    temp_file_context = request.param
    with temp_file_context("tempfile.h5", "w") as h5file:
        yield h5file


@pytest.mark.parametrize("indices", [1, slice(None), (1, slice(1, 4))])
def test_datasetslice(temp_h5file, indices):
    data = numpy.arange(50).reshape(10, 5)
    ref_data = numpy.asarray(data[indices])

    h5dataset = temp_h5file.create_group("group").create_dataset("dataset", data=data)

    with DatasetSlice(h5dataset, indices, attrs={}) as dset:
        assert silx.io.is_dataset(dset)
        assert dset.file == temp_h5file
        assert dset.shape == ref_data.shape
        assert dset.size == ref_data.size
        assert dset.dtype == ref_data.dtype
        assert len(dset) == len(ref_data)
        assert numpy.array_equal(dset[()], ref_data)
        assert dset.name == h5dataset.name


def test_datasetslice_on_external_link(tmp_path):
    data = numpy.arange(10).reshape(5, 2)

    external_filename = str(tmp_path / "external.h5")
    ext_dataset_name = "/external_data"
    with h5py.File(external_filename, "w") as h5file:
        h5file[ext_dataset_name] = data

    with h5py.File(tmp_path / "test.h5", "w") as h5file:
        h5file["group/data"] = h5py.ExternalLink(external_filename, ext_dataset_name)

        with DatasetSlice(h5file["group/data"], slice(None), attrs={}) as dset:
            assert dset.name == ext_dataset_name
            assert numpy.array_equal(dset[()], data)


@pytest.mark.parametrize(
    "shape,outer_indices,indices",
    [
        ((2, 5, 10), (-1, slice(None), slice(None)), slice(None)),
        ((2, 5, 10), (-1, slice(None), slice(None)), Ellipsis),
        # negative strides
        ((5, 10), (slice(1, 5, 2), slice(2, 8)), (slice(2, 3), slice(4, None, -2))),
        (
            (5, 10),
            (slice(4, None, -1), slice(9, 3, -2)),
            (slice(1, 3), slice(3, 0, -1)),
        ),
        ((5, 10), (slice(1, 8, 2), slice(None)), slice(2, 8)),  # slice overflow
    ],
)
def test_combine_indices(shape, outer_indices, indices):
    data = numpy.arange(numpy.prod(shape)).reshape(shape)
    ref_data = data[outer_indices][indices]

    combined_indices = _combine_indices(shape, outer_indices, indices)

    assert numpy.array_equal(data[combined_indices], ref_data)



# /*##########################################################################
#
# Copyright (c) 2015-2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module provides API to manage colors.
"""

from __future__ import annotations

__authors__ = ["T. Vincent", "H.Payno"]
__license__ = "MIT"
__date__ = "29/01/2019"


import numpy
import logging
import numbers
import re
from collections.abc import Iterable
from typing import Any, Sequence, Tuple, Union

import silx
from silx._utils import NP_OPTIONAL_COPY
from silx.gui import qt
from silx.gui.utils import blockSignals
from silx.math import colormap as _colormap
from silx.utils.exceptions import NotEditableError


_logger = logging.getLogger(__name__)

try:
    import silx.gui.utils.matplotlib  # noqa  Initalize matplotlib

    try:
        from matplotlib import colormaps as _matplotlib_colormaps
    except ImportError:  # For matplotlib < 3.5
        from matplotlib import cm as _matplotlib_cm
        from matplotlib.pyplot import colormaps as _matplotlib_colormaps
    else:
        _matplotlib_cm = None
except ImportError:
    _logger.info("matplotlib not available, only embedded colormaps available")
    _matplotlib_cm = None
    _matplotlib_colormaps = None


_COLORDICT = {}
"""Dictionary of common colors."""

_COLORDICT["b"] = _COLORDICT["blue"] = "#0000ff"
_COLORDICT["r"] = _COLORDICT["red"] = "#ff0000"
_COLORDICT["g"] = _COLORDICT["green"] = "#00ff00"
_COLORDICT["k"] = _COLORDICT["black"] = "#000000"
_COLORDICT["w"] = _COLORDICT["white"] = "#ffffff"
_COLORDICT["pink"] = "#ff66ff"
_COLORDICT["brown"] = "#a52a2a"
_COLORDICT["orange"] = "#ff9900"
_COLORDICT["violet"] = "#6600ff"
_COLORDICT["gray"] = _COLORDICT["grey"] = "#a0a0a4"
# _COLORDICT['darkGray'] = _COLORDICT['darkGrey'] = '#808080'
# _COLORDICT['lightGray'] = _COLORDICT['lightGrey'] = '#c0c0c0'
_COLORDICT["y"] = _COLORDICT["yellow"] = "#ffff00"
_COLORDICT["m"] = _COLORDICT["magenta"] = "#ff00ff"
_COLORDICT["c"] = _COLORDICT["cyan"] = "#00ffff"
_COLORDICT["darkBlue"] = "#000080"
_COLORDICT["darkRed"] = "#800000"
_COLORDICT["darkGreen"] = "#008000"
_COLORDICT["darkBrown"] = "#660000"
_COLORDICT["darkCyan"] = "#008080"
_COLORDICT["darkYellow"] = "#808000"
_COLORDICT["darkMagenta"] = "#800080"
_COLORDICT["transparent"] = "#00000000"


# FIXME: It could be nice to expose a functional API instead of that attribute
COLORDICT = _COLORDICT


DEFAULT_MIN_LIN = 0
"""Default min value if in linear normalization"""
DEFAULT_MAX_LIN = 1
"""Default max value if in linear normalization"""


_INDEXED_COLOR_PATTERN = re.compile(r"C(?P<index>[0-9]+)")


ColorType = Union[str, Sequence[numbers.Real], qt.QColor]
"""Type of :func:`rgba`'s color argument"""


RGBAColorType = Tuple[float, float, float, float]
"""Type of :func:`rgba` return value"""


def rgba(
    color: ColorType,
    colorDict: dict[str, str] | None = None,
    colors: Sequence[str] | None = None,
) -> RGBAColorType:
    """Convert different kind of color definition to a tuple (R, G, B, A) of floats.

    It supports:
    - color names: e.g., 'green'
    - color codes: '#RRGGBB' and '#RRGGBBAA'
    - indexed color names: e.g., 'C0'
    - RGB(A) sequence of uint8 in [0, 255] or float in [0, 1]
    - QColor

    :param color: The color to convert
    :param colorDict: A dictionary of color name conversion to color code
    :param colors: Sequence of colors to use for `
    :returns: RGBA colors as floats in [0., 1.]
    :raises ValueError: if the input is not a valid color
    """
    if isinstance(color, str):
        # From name
        colorFromDict = (_COLORDICT if colorDict is None else colorDict).get(color)
        if colorFromDict is not None:
            return rgba(colorFromDict, colorDict, colors)

        # From indexed color name: color{index}
        match = _INDEXED_COLOR_PATTERN.fullmatch(color)
        if match is not None:
            if colors is None:
                colors = silx.config.DEFAULT_PLOT_CURVE_COLORS
            index = int(match["index"]) % len(colors)
            return rgba(colors[index], colorDict, colors)

        # From #code
        if len(color) in (7, 9) and color[0] == "#":
            r = int(color[1:3], 16) / 255.0
            g = int(color[3:5], 16) / 255.0
            b = int(color[5:7], 16) / 255.0
            a = int(color[7:9], 16) / 255.0 if len(color) == 9 else 1.0
            return r, g, b, a

        raise ValueError(f"The string '{color}' is not a valid color")

    # From QColor
    if isinstance(color, qt.QColor):
        return rgba(color.getRgb(), colorDict, colors)

    # From array
    values = numpy.asarray(color).ravel()

    if values.dtype.kind not in "iuf":
        raise ValueError(
            f"The array color must be integer/unsigned or float. Found '{values.dtype.kind}'"
        )
    if len(values) not in (3, 4):
        raise ValueError(
            f"The array color must have 3 or 4 compound. Found '{len(values)}'"
        )

    # Convert from integers in [0, 255] to float in [0, 1]
    if values.dtype.kind in "iu":
        values = values / 255.0

    values = numpy.clip(values, 0.0, 1.0)

    if len(values) == 3:
        return values[0], values[1], values[2], 1.0
    return tuple(values)


def greyed(
    color: ColorType,
    colorDict: dict[str, str] | None = None,
) -> RGBAColorType:
    """Convert color code '#RRGGBB' and '#RRGGBBAA' to a grey color
    (R, G, B, A).

    It also supports RGB(A) from uint8 in [0, 255], float in [0, 1], and
    QColor as color argument.

    :param color: The color to convert
    :param colorDict: A dictionary of color name conversion to color code
    :returns: RGBA colors as floats in [0., 1.]
    """
    r, g, b, a = rgba(color=color, colorDict=colorDict)
    g = 0.21 * r + 0.72 * g + 0.07 * b
    return g, g, g, a


def asQColor(color: ColorType) -> qt.QColor:
    """Convert color code '#RRGGBB' and '#RRGGBBAA' to a `qt.QColor`.

    It also supports RGB(A) from uint8 in [0, 255], float in [0, 1], and
    QColor as color argument.

    :param color: The color to convert
    """
    color = rgba(color)
    return qt.QColor.fromRgbF(*color)


def cursorColorForColormap(colormapName: str) -> str:
    """Get a color suitable for overlay over a colormap.

    :param colormapName: The name of the colormap.
    :return: Name of the color.
    """
    return _colormap.get_colormap_cursor_color(colormapName)


# Colormap loader


def _registerColormapFromMatplotlib(
    name: str,
    cursor_color: str = "black",
    preferred: bool = False,
):
    if _matplotlib_cm is not None:
        colormap = _matplotlib_cm.get_cmap(name)
    else:  # matplotlib >= 3.5
        colormap = _matplotlib_colormaps[name]
    lut = colormap(numpy.linspace(0, 1, colormap.N, endpoint=True))
    colors = _colormap.array_to_rgba8888(lut)
    registerLUT(name, colors, cursor_color, preferred)


def _getColormap(name: str) -> numpy.ndarray:
    """Returns the color LUT corresponding to a colormap name
    :param name: Name of the colormap to load
    :returns: Corresponding table of colors
    :raise ValueError: If no colormap corresponds to name
    """
    name = str(name)
    try:
        return _colormap.get_colormap_lut(name)
    except ValueError:
        # Colormap is not available, try to load it from matplotlib
        _registerColormapFromMatplotlib(name, "black", False)
    return _colormap.get_colormap_lut(name)


class _Colormappable:
    """Class for objects that can be colormapped by a :class:`Colormap`

    Used by silx.gui.plot.items.core.ColormapMixIn
    """

    def _getColormapAutoscaleRange(
        self,
        colormap: Colormap | None,
    ) -> tuple[float | None, float | None]:
        """Returns the autoscale range for given colormap.

        :param colormap:
           The colormap for which to compute the autoscale range.
           If None, the default, the colormap of the item is used
        :return: (vmin, vmax) range
        """
        raise NotImplementedError("This method must be implemented in subclass")

    def getColormappedData(copy: bool = False) -> numpy.ndarray | None:
        """Returns the data used to compute the displayed colors

        :param copy: True to get a copy, False to get internal data (do not modify!).
        """
        raise NotImplementedError("This method must be implemented in subclass")


class Colormap(qt.QObject):
    """Description of a colormap

    If no `name` nor `colors` are provided, a default gray LUT is used.

    :param name: Name of the colormap
    :param colors: optional, custom colormap.
            Nx3 or Nx4 numpy array of RGB(A) colors,
            either uint8 or float in [0, 1].
            If 'name' is None, then this array is used as the colormap.
    :param normalization: Normalization: 'linear' (default) or 'log'
    :param vmin: Lower bound of the colormap or None for autoscale (default)
    :param vmax: Upper bounds of the colormap or None for autoscale (default)
    """

    LINEAR = "linear"
    """constant for linear normalization"""

    LOGARITHM = "log"
    """constant for logarithmic normalization"""

    SQRT = "sqrt"
    """constant for square root normalization"""

    GAMMA = "gamma"
    """Constant for gamma correction normalization"""

    ARCSINH = "arcsinh"
    """constant for inverse hyperbolic sine normalization"""

    _BASIC_NORMALIZATIONS = {
        LINEAR: _colormap.LinearNormalization(),
        LOGARITHM: _colormap.LogarithmicNormalization(),
        SQRT: _colormap.SqrtNormalization(),
        ARCSINH: _colormap.ArcsinhNormalization(),
    }
    """Normalizations without parameters"""

    NORMALIZATIONS = LINEAR, LOGARITHM, SQRT, GAMMA, ARCSINH
    """Tuple of managed normalizations"""

    MINMAX = "minmax"
    """constant for autoscale using min/max data range"""

    STDDEV3 = "stddev3"
    """constant for autoscale using mean +/- 3*std(data)
    with a clamp on min/max of the data"""

    PERCENTILE_1_99 = "percentile_1_99"
    """constant for autoscale using 1st and 99th percentile of data"""

    AUTOSCALE_MODES = (MINMAX, STDDEV3, PERCENTILE_1_99)
    """Tuple of managed auto scale algorithms"""

    sigChanged = qt.Signal()
    """Signal emitted when the colormap has changed."""

    _DEFAULT_NAN_COLOR = 255, 255, 255, 0

    def __init__(
        self,
        name: str | None = None,
        colors: numpy.ndarray | None = None,
        normalization: str = LINEAR,
        vmin: float | None = None,
        vmax: float | None = None,
        autoscaleMode: str = MINMAX,
    ):
        qt.QObject.__init__(self)
        self._editable = True
        self.__gamma = 2.0
        # Default NaN color: fully transparent white
        self.__nanColor = numpy.array(self._DEFAULT_NAN_COLOR, dtype=numpy.uint8)

        assert normalization in Colormap.NORMALIZATIONS
        assert autoscaleMode in Colormap.AUTOSCALE_MODES

        if normalization is Colormap.LOGARITHM:
            if (vmin is not None and vmin < 0) or (vmax is not None and vmax < 0):
                m = "Unsuported vmin (%s) and/or vmax (%s) given for a log scale."
                m += " Autoscale will be performed."
                m = m % (vmin, vmax)
                _logger.warning(m)
                vmin = None
                vmax = None

        self._name = None
        self._colors = None

        if colors is not None and name is not None:
            raise ValueError("name and colors arguments can't be set at the same time")

        if name is not None:
            self.setName(name)  # And resets colormap LUT
        elif colors is not None:
            self.setColormapLUT(colors)
        else:
            # Default colormap is grey
            self.setName("gray")

        self._normalization = str(normalization)
        self._autoscaleMode = str(autoscaleMode)
        self._vmin = float(vmin) if vmin is not None else None
        self._vmax = float(vmax) if vmax is not None else None
        self.__warnBadVmin = True
        self.__warnBadVmax = True

    def setFromColormap(self, other: Colormap):
        """Set this colormap using information from the `other` colormap.

        :param other: Colormap to use as reference.
        """
        if not self.isEditable():
            raise NotEditableError("Colormap is not editable")
        if self == other:
            return
        with blockSignals(self):
            name = other.getName()
            if name is not None:
                self.setName(name)
            else:
                self.setColormapLUT(other.getColormapLUT())
            self.setNaNColor(other.getNaNColor())
            self.setNormalization(other.getNormalization())
            self.setGammaNormalizationParameter(other.getGammaNormalizationParameter())
            self.setAutoscaleMode(other.getAutoscaleMode())
            self.setVRange(*other.getVRange())
            self.setEditable(other.isEditable())
        self.sigChanged.emit()

    def getNColors(self, nbColors: int | None = None) -> numpy.ndarray:
        """Returns N colors computed by sampling the colormap regularly.

        :param nbColors:
            The number of colors in the returned array or None for the default value.
            The default value is the size of the colormap LUT.
        :return: 2D array of uint8 of shape (nbColors, 4)
        """
        # Handle default value for nbColors
        if nbColors is None:
            return numpy.array(self._colors, copy=True)
        else:
            nbColors = int(nbColors)
            colormap = self.copy()
            colormap.setNormalization(Colormap.LINEAR)
            colormap.setVRange(vmin=0, vmax=nbColors - 1)
            colors = colormap.applyToData(numpy.arange(nbColors, dtype=numpy.int32))
            return colors

    def getName(self) -> str | None:
        """Return the name of the colormap"""
        return self._name

    def setName(self, name: str):
        """Set the name of the colormap to use.

        :param name: The name of the colormap.
            At least the following names are supported: 'gray',
            'reversed gray', 'temperature', 'red', 'green', 'blue', 'jet',
            'viridis', 'magma', 'inferno', 'plasma'.
        """
        name = str(name)
        if self._name == name:
            return
        if self.isEditable() is False:
            raise NotEditableError("Colormap is not editable")
        if name not in self.getSupportedColormaps():
            raise ValueError("Colormap name '%s' is not supported" % name)
        self._name = name
        self._colors = _getColormap(self._name)
        self.sigChanged.emit()

    def getColormapLUT(self, copy: bool = True) -> numpy.ndarray | None:
        """Return the list of colors for the colormap or None if not set.

        This returns None if the colormap was set with :meth:`setName`.
        Use :meth:`getNColors` to get the colormap LUT for any colormap.

        :param copy: If true a copy of the numpy array is provided
        :return: the list of colors for the colormap or None if not set
        """
        if self._name is None:
            return numpy.array(self._colors, copy=copy or NP_OPTIONAL_COPY)
        return None

    def setColormapLUT(self, colors: numpy.ndarray):
        """Set the colors of the colormap.

        :param colors: the colors of the LUT.
           If float, it is converted from [0, 1] to uint8 range.
           Otherwise it is casted to uint8.

        .. warning: this will set the value of name to None
        """
        if self.isEditable() is False:
            raise NotEditableError("Colormap is not editable")
        assert colors is not None

        colors = numpy.asarray(colors)
        if colors.shape == ():
            raise TypeError(
                "An array is expected for 'colors' argument. '%s' was found."
                % type(colors)
            )
        assert len(colors) != 0
        assert colors.ndim >= 2
        colors.shape = -1, colors.shape[-1]
        self._colors = _colormap.array_to_rgba8888(colors)
        self._name = None
        self.sigChanged.emit()

    def getNaNColor(self) -> qt.QColor:
        """Returns the color to use for Not-A-Number floating point value."""
        return qt.QColor(*self.__nanColor)

    def setNaNColor(self, color: ColorType):
        """Set the color to use for Not-A-Number floating point value.

        :param color: RGB(A) color to use for NaN values
        """
        color = (numpy.array(rgba(color)) * 255).astype(numpy.uint8)
        if not numpy.array_equal(self.__nanColor, color):
            self.__nanColor = color
            self.sigChanged.emit()

    def getNormalization(self) -> str:
        """Return the normalization of the colormap.

        See :meth:`setNormalization` for returned values.

        :return: the normalization of the colormap
        """
        return self._normalization

    def setNormalization(self, norm: str):
        """Set the colormap normalization.

        Accepted normalizations: 'log', 'linear', 'sqrt'

        :param norm: the norm to set
        """
        assert norm in self.NORMALIZATIONS
        if self.isEditable() is False:
            raise NotEditableError("Colormap is not editable")
        norm = str(norm)
        if norm != self._normalization:
            self._normalization = norm
            self.__warnBadVmin = True
            self.__warnBadVmax = True
            self.sigChanged.emit()

    def setGammaNormalizationParameter(self, gamma: float):
        """Set the gamma correction parameter.

        Only used for gamma correction normalization.

        :raise ValueError: If gamma is not valid
        """
        if gamma < 0.0 or not numpy.isfinite(gamma):
            raise ValueError("Gamma value not supported")
        if gamma != self.__gamma:
            self.__gamma = gamma
            self.sigChanged.emit()

    def getGammaNormalizationParameter(self) -> float:
        """Returns the gamma correction parameter value."""
        return self.__gamma

    def getAutoscaleMode(self) -> str:
        """Return the autoscale mode of the colormap ('minmax' or 'stddev3')"""
        return self._autoscaleMode

    def setAutoscaleMode(self, mode: str):
        """Set the autoscale mode: either 'minmax' or 'stddev3'

        :param mode: the mode to set
        """
        if self.isEditable() is False:
            raise NotEditableError("Colormap is not editable")
        assert mode in self.AUTOSCALE_MODES
        if mode != self._autoscaleMode:
            self._autoscaleMode = mode
            self.sigChanged.emit()

    def isAutoscale(self) -> bool:
        """Return True if both min and max are in autoscale mode"""
        return self._vmin is None and self._vmax is None

    def getVMin(self) -> float | None:
        """Return the lower bound of the colormap

        :return: the lower bound of the colormap
        """
        return self._vmin

    def setVMin(self, vmin: float | None):
        """Set the minimal value of the colormap

        :param vmin: Lower bound of the colormap or None for autoscale (initial value)
        """
        if self.isEditable() is False:
            raise NotEditableError("Colormap is not editable")
        if vmin is not None:
            if self._vmax is not None and vmin > self._vmax:
                err = "Can't set vmin because vmin >= vmax. " "vmin = %s, vmax = %s" % (
                    vmin,
                    self._vmax,
                )
                raise ValueError(err)

        if vmin != self._vmin:
            self._vmin = vmin
            self.__warnBadVmin = True
            self.sigChanged.emit()

    def getVMax(self) -> float | None:
        """Return the upper bounds of the colormap or None

        :return: the upper bounds of the colormap or None
        """
        return self._vmax

    def setVMax(self, vmax: float | None):
        """Set the maximal value of the colormap

        :param vmax: Upper bounds of the colormap or None for autoscale (initial value)
        """
        if self.isEditable() is False:
            raise NotEditableError("Colormap is not editable")
        if vmax is not None:
            if self._vmin is not None and vmax < self._vmin:
                err = "Can't set vmax because vmax <= vmin. " "vmin = %s, vmax = %s" % (
                    self._vmin,
                    vmax,
                )
                raise ValueError(err)

        if vmax != self._vmax:
            self._vmax = vmax
            self.__warnBadVmax = True
            self.sigChanged.emit()

    def isEditable(self) -> bool:
        """Return if the colormap is editable or not

        :return: editable state of the colormap
        """
        return self._editable

    def setEditable(self, editable: bool):
        """
        Set the editable state of the colormap

        :param editable: is the colormap editable
        """
        assert type(editable) is bool
        self._editable = editable
        self.sigChanged.emit()

    def _getNormalizer(self):  # TODO
        """Returns normalizer object"""
        normalization = self.getNormalization()
        if normalization == self.GAMMA:
            return _colormap.GammaNormalization(self.getGammaNormalizationParameter())
        else:
            return self._BASIC_NORMALIZATIONS[normalization]

    def _computeAutoscaleRange(self, data: numpy.ndarray):
        """Compute the data range which will be used in autoscale mode.

        :param data: The data for which to compute the range
        :return: (vmin, vmax) range
        """
        return self._getNormalizer().autoscale(data, mode=self.getAutoscaleMode())

    def getColormapRange(
        self,
        data: numpy.ndarray | _Colormappable | None = None,
    ) -> tuple[float, float]:
        """Return (vmin, vmax) the range of the colormap for the given data or item.

        :param data: The data or item to use for autoscale bounds.
        :return: (vmin, vmax) corresponding to the colormap applied to data if provided.
        """
        vmin = self._vmin
        vmax = self._vmax
        assert (
            vmin is None or vmax is None or vmin <= vmax
        )  # TODO handle this in setters

        normalizer = self._getNormalizer()

        # Handle invalid bounds as autoscale
        if vmin is not None and not normalizer.is_valid(vmin):
            if self.__warnBadVmin:
                self.__warnBadVmin = False
                _logger.info("Invalid vmin, switching to autoscale for lower bound")
            vmin = None
        if vmax is not None and not normalizer.is_valid(vmax):
            if self.__warnBadVmax:
                self.__warnBadVmax = False
                _logger.info("Invalid vmax, switching to autoscale for upper bound")
            vmax = None

        if vmin is None or vmax is None:  # Handle autoscale
            if isinstance(data, _Colormappable):
                min_, max_ = data._getColormapAutoscaleRange(self)
                # Make sure min_, max_ are not None
                min_ = normalizer.DEFAULT_RANGE[0] if min_ is None else min_
                max_ = normalizer.DEFAULT_RANGE[1] if max_ is None else max_
            else:
                min_, max_ = normalizer.autoscale(data, mode=self.getAutoscaleMode())

            if vmin is None:  # Set vmin respecting provided vmax
                vmin = min_ if vmax is None else min(min_, vmax)

            if vmax is None:
                vmax = max(max_, vmin)  # Handle max_ <= 0 for log scale

        return vmin, vmax

    def getVRange(self) -> tuple[float | None, float | None]:
        """Get the bounds of the colormap

        :returns: A tuple of 2 values for min and max. Or None instead of float
            for autoscale
        """
        return self.getVMin(), self.getVMax()

    def setVRange(self, vmin: float | None, vmax: float | None):
        """Set the bounds of the colormap

        :param vmin: Lower bound of the colormap or None for autoscale
            (default)
        :param vmax: Upper bounds of the colormap or None for autoscale
            (default)
        """
        if self.isEditable() is False:
            raise NotEditableError("Colormap is not editable")

        if (vmin is not None and not numpy.isfinite(vmin)) or (
            vmax is not None and not numpy.isfinite(vmax)
        ):
            err = (
                "Can't set vmin and vmax because vmin or vmax are not finite "
                "vmin = %s, vmax = %s" % (vmin, vmax)
            )
            raise ValueError(err)

        if vmin is not None and vmax is not None:
            if vmin > vmax:
                err = (
                    "Can't set vmin and vmax because vmin >= vmax "
                    "vmin = %s, vmax = %s" % (vmin, vmax)
                )
                raise ValueError(err)

        if self._vmin == vmin and self._vmax == vmax:
            return

        if vmin != self._vmin:
            self.__warnBadVmin = True
        self._vmin = vmin
        if vmax != self._vmax:
            self.__warnBadVmax = True
        self._vmax = vmax
        self.sigChanged.emit()

    def __getitem__(self, item: str):
        if item == "autoscale":
            return self.isAutoscale()
        elif item == "name":
            return self.getName()
        elif item == "normalization":
            return self.getNormalization()
        elif item == "vmin":
            return self.getVMin()
        elif item == "vmax":
            return self.getVMax()
        elif item == "colors":
            return self.getColormapLUT()
        elif item == "autoscaleMode":
            return self.getAutoscaleMode()
        else:
            raise KeyError(item)

    def _toDict(self) -> dict:
        """Return the equivalent colormap as a dictionary
        (old colormap representation)

        :return: the representation of the Colormap as a dictionary
        """
        return {
            "name": self._name,
            "colors": self.getColormapLUT(),
            "vmin": self._vmin,
            "vmax": self._vmax,
            "autoscale": self.isAutoscale(),
            "normalization": self.getNormalization(),
            "autoscaleMode": self.getAutoscaleMode(),
        }

    def _setFromDict(self, dic: dict):
        """Set values to the colormap from a dictionary

        :param dic: the colormap as a dictionary
        """
        if self.isEditable() is False:
            raise NotEditableError("Colormap is not editable")
        name = dic["name"] if "name" in dic else None
        colors = dic["colors"] if "colors" in dic else None
        if name is not None and colors is not None:
            if isinstance(colors, int):
                # Filter out argument which was supported but never used
                _logger.info("Unused 'colors' from colormap dictionary filterer.")
                colors = None
        vmin = dic["vmin"] if "vmin" in dic else None
        vmax = dic["vmax"] if "vmax" in dic else None
        if "normalization" in dic:
            normalization = dic["normalization"]
        else:
            warn = "Normalization not given in the dictionary, "
            warn += "set by default to " + Colormap.LINEAR
            _logger.warning(warn)
            normalization = Colormap.LINEAR

        if name is None and colors is None:
            err = "The colormap should have a name defined or a tuple of colors"
            raise ValueError(err)
        if normalization not in Colormap.NORMALIZATIONS:
            err = "Given normalization is not recognized (%s)" % normalization
            raise ValueError(err)

        autoscaleMode = dic.get("autoscaleMode", Colormap.MINMAX)
        if autoscaleMode not in Colormap.AUTOSCALE_MODES:
            err = "Given autoscale mode is not recognized (%s)" % autoscaleMode
            raise ValueError(err)

        # If autoscale, then set boundaries to None
        if dic.get("autoscale", False):
            vmin, vmax = None, None

        if name is not None:
            self.setName(name)
        else:
            self.setColormapLUT(colors)
        self._vmin = vmin
        self._vmax = vmax
        self._autoscale = True if (vmin is None and vmax is None) else False
        self._normalization = normalization
        self._autoscaleMode = autoscaleMode

        self.__warnBadVmin = True
        self.__warnBadVmax = True
        self.sigChanged.emit()

    @staticmethod
    def _fromDict(dic: dict):
        colormap = Colormap()
        colormap._setFromDict(dic)
        return colormap

    def copy(self) -> Colormap:
        """Return a copy of the Colormap."""
        colormap = Colormap(
            name=self._name,
            colors=self.getColormapLUT(),
            vmin=self._vmin,
            vmax=self._vmax,
            normalization=self.getNormalization(),
            autoscaleMode=self.getAutoscaleMode(),
        )
        colormap.setNaNColor(self.getNaNColor())
        colormap.setGammaNormalizationParameter(self.getGammaNormalizationParameter())
        colormap.setEditable(self.isEditable())
        return colormap

    def applyToData(
        self,
        data: numpy.ndarray | _Colormappable,
        reference: numpy.ndarray | _Colormappable | None = None,
    ) -> numpy.ndarray:
        """Apply the colormap to the data

        :param data:
            The data to convert or the item for which to apply the colormap.
        :param reference:
            The data or item to use as reference to compute autoscale
        """
        if reference is None:
            reference = data
        vmin, vmax = self.getColormapRange(reference)

        if isinstance(data, _Colormappable):  # Use item's data
            data = data.getColormappedData(copy=False)

        return _colormap.cmap(
            data, self._colors, vmin, vmax, self._getNormalizer(), self.__nanColor
        )

    @staticmethod
    def getSupportedColormaps() -> tuple[str, ...]:
        """Get the supported colormap names as a tuple of str.

        The list should at least contain and start by:

         ('gray', 'reversed gray', 'temperature', 'red', 'green', 'blue',
         'viridis', 'magma', 'inferno', 'plasma')
        """
        registered_colormaps = _colormap.get_registered_colormaps()
        colormaps = set(registered_colormaps)
        if _matplotlib_colormaps is not None:
            colormaps.update(_matplotlib_colormaps())

        # Put registered_colormaps first
        colormaps = tuple(
            cmap for cmap in sorted(colormaps) if cmap not in registered_colormaps
        )
        return registered_colormaps + colormaps

    def __str__(self) -> str:
        return str(self._toDict())

    def __eq__(self, other: Any):
        """Compare colormap values and not pointers"""
        if other is None:
            return False
        if not isinstance(other, Colormap):
            return False
        if self.getNormalization() != other.getNormalization():
            return False
        if self.getNormalization() == self.GAMMA:
            delta = (
                self.getGammaNormalizationParameter()
                - other.getGammaNormalizationParameter()
            )
            if abs(delta) > 0.001:
                return False
        return (
            self.getName() == other.getName()
            and self.getAutoscaleMode() == other.getAutoscaleMode()
            and self.getVMin() == other.getVMin()
            and self.getVMax() == other.getVMax()
            and numpy.array_equal(self.getColormapLUT(), other.getColormapLUT())
        )

    _SERIAL_VERSION = 3

    def restoreState(self, byteArray: qt.QByteArray) -> bool:
        """
        Read the colormap state from a QByteArray.

        :param byteArray: Stream containing the state
        :return: True if the restoration sussseed
        """
        if self.isEditable() is False:
            raise NotEditableError("Colormap is not editable")
        stream = qt.QDataStream(byteArray, qt.QIODevice.ReadOnly)

        className = stream.readQString()
        if className != self.__class__.__name__:
            _logger.warning("Classname mismatch. Found %s." % className)
            return False

        version = stream.readUInt32()
        if version not in numpy.arange(1, self._SERIAL_VERSION + 1):
            _logger.warning("Serial version mismatch. Found %d." % version)
            return False

        name = stream.readQString()
        isNull = stream.readBool()
        if not isNull:
            vmin = stream.readQVariant()
        else:
            vmin = None
        isNull = stream.readBool()
        if not isNull:
            vmax = stream.readQVariant()
        else:
            vmax = None

        normalization = stream.readQString()
        if normalization == Colormap.GAMMA:
            gamma = stream.readFloat()
        else:
            gamma = None

        if version == 1:
            autoscaleMode = Colormap.MINMAX
        else:
            autoscaleMode = stream.readQString()

        if version <= 2:
            nanColor = self._DEFAULT_NAN_COLOR
        else:
            nanColor = (
                stream.readInt32(),
                stream.readInt32(),
                stream.readInt32(),
                stream.readInt32(),
            )

        # emit change event only once
        old = self.blockSignals(True)
        try:
            self.setName(name)
            self.setNormalization(normalization)
            self.setAutoscaleMode(autoscaleMode)
            self.setVRange(vmin, vmax)
            if gamma is not None:
                self.setGammaNormalizationParameter(gamma)
            self.setNaNColor(nanColor)
        finally:
            self.blockSignals(old)
        self.sigChanged.emit()
        return True

    def saveState(self) -> qt.QByteArray:
        """Save state of the colomap into a QDataStream."""
        data = qt.QByteArray()
        stream = qt.QDataStream(data, qt.QIODevice.WriteOnly)

        stream.writeQString(self.__class__.__name__)
        stream.writeUInt32(self._SERIAL_VERSION)
        stream.writeQString(self.getName())
        stream.writeBool(self.getVMin() is None)
        if self.getVMin() is not None:
            stream.writeQVariant(self.getVMin())
        stream.writeBool(self.getVMax() is None)
        if self.getVMax() is not None:
            stream.writeQVariant(self.getVMax())
        stream.writeQString(self.getNormalization())
        if self.getNormalization() == Colormap.GAMMA:
            stream.writeFloat(self.getGammaNormalizationParameter())
        stream.writeQString(self.getAutoscaleMode())
        nanColor = self.getNaNColor()
        stream.writeInt32(nanColor.red())
        stream.writeInt32(nanColor.green())
        stream.writeInt32(nanColor.blue())
        stream.writeInt32(nanColor.alpha())

        return data


_PREFERRED_COLORMAPS = None
"""
Tuple of preferred colormap names accessed with :meth:`preferredColormaps`.
"""

_DEFAULT_PREFERRED_COLORMAPS = (
    "gray",
    "reversed gray",
    "red",
    "green",
    "blue",
    "viridis",
    "cividis",
    "magma",
    "inferno",
    "plasma",
    "temperature",
    "jet",
    "hsv",
)


def preferredColormaps() -> tuple[str, ...]:
    """Returns the name of the preferred colormaps.

    This list is used by widgets allowing to change the colormap
    like the :class:`ColormapDialog` as a subset of colormap choices.
    """
    global _PREFERRED_COLORMAPS
    if _PREFERRED_COLORMAPS is None:
        # Initialize preferred colormaps
        setPreferredColormaps(_DEFAULT_PREFERRED_COLORMAPS)
    return tuple(_PREFERRED_COLORMAPS)


def setPreferredColormaps(colormaps: Iterable[str]):
    """Set the list of preferred colormap names.

    Warning: If a colormap name is not available
    it will be removed from the list.

    :param colormaps: Not empty list of colormap names
    :raise ValueError: if the list of available preferred colormaps is empty.
    """
    supportedColormaps = Colormap.getSupportedColormaps()
    colormaps = [cmap for cmap in colormaps if cmap in supportedColormaps]
    if len(colormaps) == 0:
        raise ValueError("Cannot set preferred colormaps to an empty list")

    global _PREFERRED_COLORMAPS
    _PREFERRED_COLORMAPS = colormaps


def registerLUT(
    name: str,
    colors: numpy.ndarray,
    cursor_color: str = "black",
    preferred: bool = True,
):
    """Register a custom LUT to be used with `Colormap` objects.

    It can override existing LUT names.

    :param name: Name of the LUT as defined to configure colormaps
    :param colors: The custom LUT to register.
            Nx3 or Nx4 numpy array of RGB(A) colors,
            either uint8 or float in [0, 1].
    :param preferred: If true, this LUT will be displayed as part of the
        preferred colormaps in dialogs.
    :param cursor_color: Color used to display overlay over images using
        colormap with this LUT.
    """
    _colormap.register_colormap(name, colors, cursor_color)

    if preferred:
        # Invalidate the preferred cache
        global _PREFERRED_COLORMAPS
        if _PREFERRED_COLORMAPS is not None:
            if name not in _PREFERRED_COLORMAPS:
                _PREFERRED_COLORMAPS.append(name)
        else:
            # The cache is not yet loaded, it's fine
            pass


# Load some colormaps from matplotlib by default
if _matplotlib_cm is not None:
    _registerColormapFromMatplotlib("jet", cursor_color="pink", preferred=True)
    _registerColormapFromMatplotlib("hsv", cursor_color="black", preferred=True)



# /*##########################################################################
#
# Copyright (c) 2024 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
:mod:`silx.gui.plot.actions.image` provides a set of QAction relative to data processing
and outputs for a :class:`.PlotWidget`.

The following QAction are available:

- :class:`AggregationModeAction`
"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "19/09/2024"

import logging
from silx.gui import qt
from silx.gui import icons
from ..items.image_aggregated import ImageDataAggregated

_logger = logging.getLogger(__name__)

class AggregationModeAction(qt.QWidgetAction):
    """Action providing filters for an aggregated image"""

    sigAggregationModeChanged = qt.Signal()
    """Signal emitted when the aggregation mode has changed"""  

    def __init__(self, parent):
        qt.QWidgetAction.__init__(self, parent)

        toolButton = qt.QToolButton(parent)

        filterAction = qt.QAction(self)
        filterAction.setText("No filter")
        filterAction.setCheckable(True)
        filterAction.setChecked(True)
        filterAction.setProperty(
            "aggregation", ImageDataAggregated.Aggregation.NONE
        )
        densityNoFilterAction = filterAction

        filterAction = qt.QAction(self)
        filterAction.setText("Max filter")
        filterAction.setCheckable(True)
        filterAction.setProperty(
            "aggregation", ImageDataAggregated.Aggregation.MAX
        )
        densityMaxFilterAction = filterAction

        filterAction = qt.QAction(self)
        filterAction.setText("Mean filter")
        filterAction.setCheckable(True)
        filterAction.setProperty(
            "aggregation", ImageDataAggregated.Aggregation.MEAN
        )
        densityMeanFilterAction = filterAction

        filterAction = qt.QAction(self)
        filterAction.setText("Min filter")
        filterAction.setCheckable(True)
        filterAction.setProperty(
            "aggregation", ImageDataAggregated.Aggregation.MIN
        )
        densityMinFilterAction = filterAction

        densityGroup = qt.QActionGroup(self)
        densityGroup.setExclusive(True)
        densityGroup.addAction(densityNoFilterAction)
        densityGroup.addAction(densityMaxFilterAction)
        densityGroup.addAction(densityMeanFilterAction)
        densityGroup.addAction(densityMinFilterAction)
        densityGroup.triggered.connect(self._aggregationModeChanged)
        self.__densityGroup = densityGroup

        filterMenu = qt.QMenu(toolButton)
        filterMenu.addAction(densityNoFilterAction)
        filterMenu.addAction(densityMaxFilterAction)
        filterMenu.addAction(densityMeanFilterAction)
        filterMenu.addAction(densityMinFilterAction)

        toolButton.setPopupMode(qt.QToolButton.InstantPopup)
        toolButton.setMenu(filterMenu)
        toolButton.setText("Data filters")
        toolButton.setToolTip("Enable/disable filter on the image")
        icon = icons.getQIcon("aggregation-mode")
        toolButton.setIcon(icon)
        toolButton.setText("Pixel aggregation filter")

        self.setDefaultWidget(toolButton)

    def _aggregationModeChanged(self):
        self.sigAggregationModeChanged.emit()

    def setAggregationMode(self, mode):
        """Set an Aggregated enum from ImageDataAggregated"""
        for a in self.__densityGroup.actions():
            if a.property("aggregation") is mode:
                a.setChecked(True)

    def getAggregationMode(self):
        """Returns an Aggregated enum from ImageDataAggregated"""
        densityAction = self.__densityGroup.checkedAction()
        if densityAction is None:
            return ImageDataAggregated.Aggregation.NONE
        return densityAction.property("aggregation")



import os
import pytest
import h5py
import numpy
from silx.io import open
from silx.io import h5link_utils


@pytest.fixture(scope="module")
def hdf5_with_external_data(tmpdir_factory):
    tmpdir = tmpdir_factory.mktemp("hdf5_with_external_data")
    master = str(tmpdir / "master.h5")
    external_h5 = str(tmpdir / "external.h5")
    external_raw = str(tmpdir / "external.raw")

    data = numpy.array([100, 1000, 10000], numpy.uint16)
    tshape = (1,) + data.shape

    with h5py.File(master, "w") as fmaster:
        dset = fmaster.create_dataset("data", data=data)

        fmaster["int"] = h5py.SoftLink("data")

        layout = h5py.VirtualLayout(shape=tshape, dtype=data.dtype)
        layout[0] = h5py.VirtualSource(".", "data", shape=data.shape)
        fmaster.create_virtual_dataset("vds0", layout)

        with h5py.File(external_h5, "w") as f:
            dset = f.create_dataset("data", data=data)
            layout = h5py.VirtualLayout(shape=tshape, dtype=data.dtype)
            layout[0] = h5py.VirtualSource(dset)
            fmaster.create_virtual_dataset("vds1", layout)

            layout = h5py.VirtualLayout(shape=tshape, dtype=data.dtype)
            layout[0] = h5py.VirtualSource(
                external_h5,
                "data",
                shape=data.shape,
            )
            fmaster.create_virtual_dataset("vds2", layout)
            fmaster["ext1"] = h5py.ExternalLink(external_h5, "data")

            layout = h5py.VirtualLayout(shape=tshape, dtype=data.dtype)
            layout[0] = h5py.VirtualSource(
                "external.h5",
                "data",
                shape=data.shape,
            )
            fmaster.create_virtual_dataset("vds3", layout)
            fmaster["ext2"] = h5py.ExternalLink("external.h5", "data")

            layout = h5py.VirtualLayout(shape=tshape, dtype=data.dtype)
            layout[0] = h5py.VirtualSource(
                "./external.h5",
                "data",
                shape=data.shape,
            )
            fmaster.create_virtual_dataset("vds4", layout)
            fmaster["ext3"] = h5py.ExternalLink("./external.h5", "data")

        data.tofile(external_raw)

        external = [(external_raw, 0, 16 * 3)]
        fmaster.create_dataset(
            "raw1", external=external, shape=tshape, dtype=data.dtype
        )

        external = [("external.raw", 0, 16 * 3)]
        fmaster.create_dataset(
            "raw2", external=external, shape=tshape, dtype=data.dtype
        )

        external = [("./external.raw", 0, 16 * 3)]
        fmaster.create_dataset(
            "raw3", external=external, shape=tshape, dtype=data.dtype
        )

    # Validate links
    expected = data.tolist()
    cwd = os.getcwd()
    with h5py.File(master, "r") as master:
        for name in master:
            if name in ("raw2", "raw3"):
                os.chdir(str(tmpdir))
            try:
                data = master[name][()].flatten().tolist()
            except Exception:
                assert False, name
            finally:
                if name in ("raw2", "raw3"):
                    os.chdir(cwd)
            assert data == expected, name

    return tmpdir


@pytest.mark.skipif("VirtualLayout" not in dir(h5py), reason="h5py is too old")
def test_external_dataset_info(hdf5_with_external_data):
    tmpdir = hdf5_with_external_data
    master = str(tmpdir / "master.h5")
    external_h5 = str(tmpdir / "external.h5")
    external_raw = str(tmpdir / "external.raw")
    with open(master) as f:
        for name in f:
            hdf5obj = f[name]
            info = h5link_utils.external_dataset_info(hdf5obj)
            if name in ("data", "int", "ext1", "ext2", "ext3"):
                assert info is None, name
            elif name == "vds0":
                assert info.first_source_url == f"{master}::/data"
            elif name in ("vds1", "vds2", "vds3", "vds4"):
                assert info.first_source_url == f"{external_h5}::/data"
            elif name in ("raw1", "raw2", "raw3"):
                assert info.first_source_url == external_raw
            else:
                assert False, name



# /*##########################################################################
#
# Copyright (c) 2020-2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module provides the :func:`isOpenGLAvailable` utility function.
"""
from __future__ import annotations


import os
import sys
import subprocess
from silx.gui import qt


class _isOpenGLAvailableResult:
    """Store result of checking OpenGL availability.

    It provides a `status` boolean attribute storing the result of the check and
    an `error` string attribute storting the possible error message.
    """

    def __init__(self, error: str = "", status: bool = False):
        self.__error = str(error)
        self.__status = bool(status)

    status = property(lambda self: self.__status, doc="True if OpenGL is working")
    error = property(lambda self: self.__error, doc="Error message")

    def __bool__(self):
        return self.status

    def __repr__(self):
        return f'<_isOpenGLAvailableResult: {self.status}, "{self.error}">'


def _runtimeOpenGLCheck(
    version: tuple[int, int],
    shareOpenGLContexts: bool,
) -> _isOpenGLAvailableResult:
    """Run OpenGL check in a subprocess.

    This is done by starting a subprocess that displays a Qt OpenGL widget.

    :param version:
        The minimal required OpenGL version as a 2-tuple (major, minor).
    :param shareOpenGLContexts:
        True to test the `QApplication` with `AA_ShareOpenGLContexts`.
    :return: Result status and error message
    """
    major, minor = str(version[0]), str(version[1])
    env = os.environ.copy()
    env["PYTHONPATH"] = os.pathsep.join([os.path.abspath(p) for p in sys.path])

    cmd = [sys.executable, "-s", "-S", __file__, major, minor]
    if shareOpenGLContexts:
        cmd.append("--shareOpenGLContexts")

    try:
        output = subprocess.check_output(cmd, env=env, timeout=2)
    except subprocess.TimeoutExpired:
        error = "Qt OpenGL widget hang"
        if sys.platform.startswith("linux"):
            error += ":\nIf connected remotely, GLX forwarding might be disabled."
        return _isOpenGLAvailableResult(error)
    except subprocess.CalledProcessError as e:
        return _isOpenGLAvailableResult(
            f"Qt OpenGL widget error: retcode={e.returncode}, error={e.output}"
        )

    return _isOpenGLAvailableResult(output.decode(), status=True)


_runtimeCheckCache = {}  # Cache runtime check results: {version: result}


def isOpenGLAvailable(
    version: tuple[int, int] = (2, 1),
    runtimeCheck: bool = True,
    shareOpenGLContexts: bool = False,
) -> _isOpenGLAvailableResult:
    """Check if OpenGL is available through Qt and actually working.

    After some basic tests, this is done by starting a subprocess that
    displays a Qt OpenGL widget.

    :param version:
        The minimal required OpenGL version as a 2-tuple (major, minor).
        Default: (2, 1)
    :param runtimeCheck:
        True (default) to run the test creating a Qt OpenGL widget in a subprocess,
        False to avoid this check.
    :param shareOpenGLContexts:
        True to test the `QApplication` with `AA_ShareOpenGLContexts`.
        This only can be checked with `runtimeCheck` enabled.
        Default is false.
    :return: A result object that evaluates to True if successful and
        which has a `status` boolean attribute (True if successful) and
        an `error` string attribute that is not empty if `status` is False.
    """
    if sys.platform.startswith("linux") and not os.environ.get("DISPLAY", ""):
        # On Linux and no DISPLAY available (e.g., ssh without -X)
        return _isOpenGLAvailableResult("DISPLAY environment variable not set")

    # Check pyopengl availability
    try:
        from silx.gui._glutils import gl
    except ImportError:
        return _isOpenGLAvailableResult(
            "Cannot import OpenGL wrapper: pyopengl is not installed"
        )

    # Pre checks for Qt < 5.4
    if not hasattr(qt, "QOpenGLWidget"):
        if not qt.HAS_OPENGL:
            return _isOpenGLAvailableResult(f"{qt.BINDING}.QtOpenGL not available")

        if (
            qt.BINDING == "PyQt5"
            and qt.QApplication.instance()
            and not qt.QGLFormat.hasOpenGL()
        ):
            # qt.QGLFormat.hasOpenGL MUST be called with a QApplication created
            # so this is only checked if the QApplication is already created
            return _isOpenGLAvailableResult("Qt reports OpenGL not available")

    # Check compatibility between Qt platform and pyopengl selected platform
    qt_qpa_platform = qt.QGuiApplication.platformName()
    pyopengl_platform = gl.getPlatform()
    if (qt_qpa_platform == "wayland" and pyopengl_platform != "EGLPlatform") or (
        qt_qpa_platform == "xcb" and pyopengl_platform != "GLXPlatform"
    ):
        return _isOpenGLAvailableResult(
            f"Qt platform '{qt_qpa_platform}' is not compatible with PyOpenGL platform '{pyopengl_platform}'"
        )

    keyCache = version, shareOpenGLContexts
    if keyCache in _runtimeCheckCache:  # Use cache
        return _runtimeCheckCache[keyCache]

    if not runtimeCheck:
        return _isOpenGLAvailableResult(status=True)

    # Run test in subprocess
    result = _runtimeOpenGLCheck(version, shareOpenGLContexts)
    _runtimeCheckCache[keyCache] = result
    return result


if __name__ == "__main__":
    from silx.gui._glutils import OpenGLWidget
    from silx.gui._glutils import gl
    import argparse

    class _TestOpenGLWidget(OpenGLWidget):
        """Widget checking that OpenGL is indeed available

        :param version: (major, minor) minimum OpenGL version
        """

        def __init__(self, version: tuple[int, int]):
            super(_TestOpenGLWidget, self).__init__(
                alphaBufferSize=0,
                depthBufferSize=0,
                stencilBufferSize=0,
                version=version,
            )

        def paintEvent(self, event):
            super(_TestOpenGLWidget, self).paintEvent(event)

            # Check once paint has been done
            app = qt.QApplication.instance()
            if not self.isValid():
                print("OpenGL widget is not valid")
                app.exit(1)
            else:
                qt.QTimer.singleShot(100, app.quit)

        def paintGL(self):
            gl.glClearColor(1.0, 0.0, 0.0, 0.0)
            gl.glClear(gl.GL_COLOR_BUFFER_BIT)

    parser = argparse.ArgumentParser()
    parser.add_argument("major")
    parser.add_argument("minor")
    parser.add_argument("--shareOpenGLContexts", action="store_true")

    args = parser.parse_args(args=sys.argv[1:])

    if args.shareOpenGLContexts:
        qt.QCoreApplication.setAttribute(qt.Qt.AA_ShareOpenGLContexts)
    app = qt.QApplication([])
    window = qt.QMainWindow(
        flags=qt.Qt.Popup
        | qt.Qt.FramelessWindowHint
        | qt.Qt.NoDropShadowWindowHint
        | qt.Qt.WindowStaysOnTopHint
    )
    window.setAttribute(qt.Qt.WA_ShowWithoutActivating)
    window.move(0, 0)
    window.resize(3, 3)
    widget = _TestOpenGLWidget(version=(args.major, args.minor))
    window.setCentralWidget(widget)
    window.setWindowOpacity(0.04)
    window.show()

    qt.QTimer.singleShot(1000, app.quit)
    sys.exit(app.exec())



import pytest
from silx.gui import qt
from silx.gui.widgets.WaitingOverlay import WaitingOverlay
from silx.gui.plot import Plot2D
from silx.gui.plot.PlotWidget import PlotWidget


@pytest.mark.parametrize("widget_parent", (Plot2D, qt.QFrame))
def test_show(qapp, qapp_utils, widget_parent):
    """Simple test of the WaitingOverlay component"""
    widget = widget_parent()
    widget.setAttribute(qt.Qt.WA_DeleteOnClose)

    waitingOverlay = WaitingOverlay(widget)
    waitingOverlay.setAttribute(qt.Qt.WA_DeleteOnClose)

    widget.show()
    qapp_utils.qWaitForWindowExposed(widget)
    assert waitingOverlay._waitingButton.isWaiting()

    waitingOverlay.setText("test")
    qapp.processEvents()
    assert waitingOverlay.text() == "test"
    qapp_utils.qWait(1000)

    waitingOverlay.hide()
    qapp.processEvents()

    widget.close()
    waitingOverlay.close()
    qapp.processEvents()



# /*##########################################################################
# Copyright (C) 2018-2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/

import pytest
import h5py
from ..parseutils import filenames_to_dataurls


@pytest.fixture(scope="module")
def data_path(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("silx_app_utils")
    with h5py.File(tmp_path / "test1.h5", "w") as h5:
        h5["g1/sub1/data1"] = 1
        h5["g1/sub1/data2"] = 2
        h5["g1/sub2/data1"] = 3
    return tmp_path


def test_h5__datapath(data_path):
    urls = filenames_to_dataurls([data_path / "test1.h5::g1/sub1/data1"])
    urls = list(urls)
    assert len(urls) == 1
    assert urls[0].data_path().replace("\\", "/") == "g1/sub1/data1"


def test_h5__datapath_not_existing(data_path):
    urls = filenames_to_dataurls([data_path / "test1.h5::g1/sub0/data1"])
    urls = list(urls)
    assert len(urls) == 1
    assert urls[0].data_path().replace("\\", "/") == "g1/sub0/data1"


def test_h5__datapath_with_magic(data_path):
    urls = filenames_to_dataurls([data_path / "test1.h5::g1/sub*/data*"])
    urls = list(urls)
    assert len(urls) == 3


def test_h5__datapath_with_magic_not_existing(data_path):
    urls = filenames_to_dataurls([data_path / "test1.h5::g1/sub0/data*"])
    urls = list(urls)
    assert len(urls) == 0


def test_h5__datapath_with_recursive_magic(data_path):
    urls = filenames_to_dataurls([data_path / "test1.h5::**/data1"])
    urls = list(urls)
    assert len(urls) == 2



# /*##########################################################################
#
# Copyright (c) 2004-2022 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Load Qt binding"""

__authors__ = ["V.A. Sole"]
__license__ = "MIT"
__date__ = "12/01/2022"


import importlib
import logging
import os
import sys
import traceback

from packaging.version import Version
from silx.utils import deprecation

_logger = logging.getLogger(__name__)


BINDING = None
"""The name of the Qt binding in use: PyQt5, PySide6, PyQt6."""

QtBinding = None  # noqa
"""The Qt binding module in use: PyQt5, PySide6, PyQt6."""

HAS_SVG = False
"""True if Qt provides support for Scalable Vector Graphics (QtSVG)."""

HAS_OPENGL = False
"""True if Qt provides support for OpenGL (QtOpenGL)."""


def _select_binding() -> str:
    """Select and load a Qt binding

    Qt binding is selected according to:
    - Already loaded binding
    - QT_API environment variable
    - Bindings order of priority

    :raises ImportError:
    :returns: Loaded binding
    """
    bindings = "PyQt5", "PySide6", "PyQt6"

    envvar = os.environ.get("QT_API", "").lower()

    # First check for an already loaded binding
    for binding in bindings:
        if f"{binding}.QtCore" in sys.modules:
            if envvar and envvar != binding.lower():
                _logger.warning(
                    f"Cannot satisfy QT_API={envvar} environment variable, {binding} is already loaded"
                )
            return binding

    # Check if QT_API can be satisfied
    if envvar:
        selection = [b for b in bindings if envvar == b.lower()]
        if not selection:
            _logger.warning(f"Environment variable QT_API={envvar} is not supported")
        else:
            binding = selection[0]
            try:
                importlib.import_module(f"{binding}.QtCore")
            except ImportError:
                _logger.warning(
                    f"Cannot import {binding} specified by QT_API environment variable"
                )
            else:
                return binding

    # Try to load binding
    for binding in bindings:
        try:
            importlib.import_module(f"{binding}.QtCore")
        except ImportError:
            if binding in sys.modules:
                del sys.modules[binding]
        else:
            return binding

    raise ImportError("No Qt wrapper found. Install PyQt5, PySide6, PyQt6.")


BINDING = _select_binding()


if BINDING == "PyQt5":
    _logger.debug("Using PyQt5 bindings")
    from PyQt5 import QtCore

    if sys.version_info >= (3, 10) and QtCore.PYQT_VERSION < 0x50E02:
        raise RuntimeError(
            "PyQt5 v%s is not supported, please upgrade it." % QtCore.PYQT_VERSION_STR
        )

    import PyQt5 as QtBinding  # noqa

    from PyQt5.QtCore import *  # noqa
    from PyQt5.QtGui import *  # noqa
    from PyQt5.QtWidgets import *  # noqa
    from PyQt5.QtPrintSupport import *  # noqa

    try:
        from PyQt5.QtOpenGL import *  # noqa
    except ImportError:
        _logger.info("PyQt5.QtOpenGL not available")
        HAS_OPENGL = False
    else:
        HAS_OPENGL = True

    try:
        from PyQt5.QtSvg import *  # noqa
    except ImportError:
        _logger.info("PyQt5.QtSvg not available")
        HAS_SVG = False
    else:
        HAS_SVG = True

    from PyQt5.uic import loadUi  # noqa

    Signal = pyqtSignal

    Property = pyqtProperty

    Slot = pyqtSlot

    # Disable PyQt5's cooperative multi-inheritance since other bindings do not provide it.
    # See https://www.riverbankcomputing.com/static/Docs/PyQt5/multiinheritance.html?highlight=inheritance
    class _Foo(object):
        pass

    class QObject(QObject, _Foo):
        pass

elif BINDING == "PySide6":
    _logger.debug("Using PySide6 bindings")

    import PySide6 as QtBinding  # noqa

    if Version(QtBinding.__version__) < Version("6.4"):
        raise RuntimeError(
            f"PySide6 v{QtBinding.__version__} is not supported, please upgrade it."
        )

    from PySide6.QtCore import *  # noqa
    from PySide6.QtGui import *  # noqa
    from PySide6.QtWidgets import *  # noqa
    from PySide6.QtPrintSupport import *  # noqa

    try:
        from PySide6.QtOpenGL import *  # noqa
        from PySide6.QtOpenGLWidgets import QOpenGLWidget  # noqa
    except ImportError:
        _logger.info("PySide6's QtOpenGL or QtOpenGLWidgets not available")
        HAS_OPENGL = False
    else:
        HAS_OPENGL = True

    try:
        from PySide6.QtSvg import *  # noqa
    except ImportError:
        _logger.info("PySide6.QtSvg not available")
        HAS_SVG = False
    else:
        HAS_SVG = True

    pyqtSignal = Signal


elif BINDING == "PyQt6":
    _logger.debug("Using PyQt6 bindings")

    # Monkey-patch module to expose enum values for compatibility
    # All Qt modules loaded here should be patched.
    from . import _pyqt6
    from PyQt6 import QtCore

    if QtCore.PYQT_VERSION < int("0x60300", 16):
        raise RuntimeError(
            "PyQt6 v%s is not supported, please upgrade it." % QtCore.PYQT_VERSION_STR
        )

    from PyQt6 import QtGui, QtWidgets, QtPrintSupport, QtOpenGL, QtSvg
    from PyQt6 import QtTest as _QtTest

    _pyqt6.patch_enums(
        QtCore, QtGui, QtWidgets, QtPrintSupport, QtOpenGL, QtSvg, _QtTest
    )

    import PyQt6 as QtBinding  # noqa

    from PyQt6.QtCore import *  # noqa
    from PyQt6.QtGui import *  # noqa
    from PyQt6.QtWidgets import *  # noqa
    from PyQt6.QtPrintSupport import *  # noqa

    try:
        from PyQt6.QtOpenGL import *  # noqa
        from PyQt6.QtOpenGLWidgets import QOpenGLWidget  # noqa
    except ImportError:
        _logger.info("PyQt6's QtOpenGL or QtOpenGLWidgets not available")
        HAS_OPENGL = False
    else:
        HAS_OPENGL = True

    try:
        from PyQt6.QtSvg import *  # noqa
    except ImportError:
        _logger.info("PyQt6.QtSvg not available")
        HAS_SVG = False
    else:
        HAS_SVG = True

    from PyQt6.uic import loadUi  # noqa

    Signal = pyqtSignal

    Property = pyqtProperty

    Slot = pyqtSlot

    # Disable PyQt6 cooperative multi-inheritance since other bindings do not provide it.
    # See https://www.riverbankcomputing.com/static/Docs/PyQt6/multiinheritance.html?highlight=inheritance
    class _Foo(object):
        pass

    class QObject(QObject, _Foo):
        pass

else:
    raise ImportError("No Qt wrapper found. Install PyQt5, PySide6 or PyQt6")


# provide a exception handler but not implement it by default
def exceptionHandler(type_, value, trace):
    """
    This exception handler prevents quitting to the command line when there is
    an unhandled exception while processing a Qt signal.

    The script/application willing to use it should implement code similar to:

    .. code-block:: python

        if __name__ == "__main__":
            sys.excepthook = qt.exceptionHandler

    """
    _logger.error("%s %s %s", type_, value, "".join(traceback.format_tb(trace)))
    msg = QMessageBox()
    msg.setWindowTitle("Unhandled exception")
    msg.setIcon(QMessageBox.Critical)
    msg.setInformativeText("%s %s\nPlease report details" % (type_, value))
    msg.setDetailedText(("%s " % value) + "".join(traceback.format_tb(trace)))
    msg.raise_()
    msg.exec()



# /*##########################################################################
#
# Copyright (c) 2004-2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
:mod:`silx.gui.plot.actions.fit` module provides actions relative to fit.

The following QAction are available:

- :class:`.FitAction`

.. autoclass:`.FitAction`
"""

__authors__ = ["V.A. Sole", "T. Vincent", "P. Knobel"]
__license__ = "MIT"
__date__ = "10/10/2018"

import logging
import sys
import weakref
import numpy

from .PlotToolAction import PlotToolAction
from .. import items
from silx._utils import NP_OPTIONAL_COPY
from silx.gui import qt
from silx.gui.plot.ItemsSelectionDialog import ItemsSelectionDialog

_logger = logging.getLogger(__name__)


def _getUniqueCurveOrHistogram(plot):
    """Returns unique :class:`Curve` or :class:`Histogram` in a `PlotWidget`.

    If there is an active curve, returns it, else return curve or histogram
    only if alone in the plot.

    :param PlotWidget plot:
    :rtype: Union[None,~silx.gui.plot.items.Curve,~silx.gui.plot.items.Histogram]
    """
    curve = plot.getActiveCurve()
    if curve is not None:
        return curve

    visibleItems = [item for item in plot.getItems() if item.isVisible()]
    histograms = [item for item in visibleItems if isinstance(item, items.Histogram)]
    curves = [item for item in visibleItems if isinstance(item, items.Curve)]

    if len(histograms) == 1 and len(curves) == 0:
        return histograms[0]
    elif len(curves) == 1 and len(histograms) == 0:
        return curves[0]
    else:
        return None


class _FitItemSelector(qt.QObject):
    """
    :class:`PlotWidget` observer that emits signal when fit selection changes.

    Track active curve or unique curve or histogram.
    """

    sigCurrentItemChanged = qt.Signal(object)
    """Signal emitted when the item to fit has changed"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__plotWidgetRef = None
        self.__currentItem = None

    def getCurrentItem(self):
        """Return currently selected item

        :rtype: Union[Item,None]
        """
        return self.__currentItem

    def getPlotWidget(self):
        """Return currently attached :class:`PlotWidget`

        :rtype: Union[PlotWidget,None]
        """
        return None if self.__plotWidgetRef is None else self.__plotWidgetRef()

    def setPlotWidget(self, plotWidget):
        """Set the :class:`PlotWidget` for which to track changes

        :param Union[PlotWidget,None] plotWidget:
            The :class:`PlotWidget` to observe
        """
        # disconnect from previous plot
        previousPlotWidget = self.getPlotWidget()
        if previousPlotWidget is not None:
            previousPlotWidget.sigItemAdded.disconnect(self.__plotWidgetUpdated)
            previousPlotWidget.sigItemRemoved.disconnect(self.__plotWidgetUpdated)
            previousPlotWidget.sigActiveCurveChanged.disconnect(
                self.__plotWidgetUpdated
            )

        if plotWidget is None:
            self.__plotWidgetRef = None
            self.__setCurrentItem(None)
            return
        self.__plotWidgetRef = weakref.ref(plotWidget, self.__plotDeleted)

        # connect to new plot
        plotWidget.sigItemAdded.connect(self.__plotWidgetUpdated)
        plotWidget.sigItemRemoved.connect(self.__plotWidgetUpdated)
        plotWidget.sigActiveCurveChanged.connect(self.__plotWidgetUpdated)
        self.__plotWidgetUpdated()

    def __plotDeleted(self):
        """Handle deletion of PlotWidget"""
        self.__setCurrentItem(None)

    def __plotWidgetUpdated(self, *args, **kwargs):
        """Handle updates of PlotWidget content"""
        plotWidget = self.getPlotWidget()
        if plotWidget is None:
            return
        self.__setCurrentItem(_getUniqueCurveOrHistogram(plotWidget))

    def __setCurrentItem(self, item):
        """Handle change of current item"""
        if sys.is_finalizing():
            return

        previousItem = self.getCurrentItem()
        if item != previousItem:
            if previousItem is not None:
                previousItem.sigItemChanged.disconnect(self.__itemUpdated)

            self.__currentItem = item

            if self.__currentItem is not None:
                self.__currentItem.sigItemChanged.connect(self.__itemUpdated)
            self.sigCurrentItemChanged.emit(self.__currentItem)

    def __itemUpdated(self, event):
        """Handle change on current item"""
        if event == items.ItemChangedType.DATA:
            self.sigCurrentItemChanged.emit(self.__currentItem)


class FitAction(PlotToolAction):
    """QAction to open a :class:`FitWidget` and set its data to the
    active curve if any, or to the first curve.

    :param plot: :class:`.PlotWidget` instance on which to operate
    :param parent: See :class:`QAction`
    """

    def __init__(self, plot, parent=None):
        self.__item = None
        self.__activeCurveSynchroEnabled = False
        self.__range = 0, 1
        self.__rangeAutoUpdate = False
        self.__x, self.__y = None, None  # Data to fit
        self.__curveParams = {}  # Store curve parameters to use for fit result
        self.__legend = None

        super(FitAction, self).__init__(
            plot,
            icon="math-fit",
            text="Fit curve",
            tooltip="Open a fit dialog",
            parent=parent,
        )

        self.__fitItemSelector = _FitItemSelector()
        self.__fitItemSelector.sigCurrentItemChanged.connect(self._setFittedItem)

    def _createToolWindow(self):
        # import done here rather than at module level to avoid circular import
        # FitWidget -> BackgroundWidget -> PlotWindow -> actions -> fit -> FitWidget
        from ...fit.FitWidget import FitWidget

        window = FitWidget(parent=self.plot)
        window.setWindowFlags(qt.Qt.Dialog)
        window.sigFitWidgetSignal.connect(self.handle_signal)
        return window

    def _connectPlot(self, window):
        if self.isXRangeUpdatedOnZoom():
            self.__setAutoXRangeEnabled(True)
        else:
            plot = self.plot
            if plot is None:
                _logger.error("No associated PlotWidget")
                return
            self._setXRange(*plot.getXAxis().getLimits())

        if self.isFittedItemUpdatedFromActiveCurve():
            self.__setFittedItemAutoUpdateEnabled(True)
        else:
            # Wait for the next iteration, else the plot is not yet initialized
            # No curve available
            qt.QTimer.singleShot(10, self._initFit)

    def _disconnectPlot(self, window):
        if self.isXRangeUpdatedOnZoom():
            self.__setAutoXRangeEnabled(False)

        if self.isFittedItemUpdatedFromActiveCurve():
            self.__setFittedItemAutoUpdateEnabled(False)

    def _initFit(self):
        plot = self.plot
        if plot is None:
            _logger.error("No associated PlotWidget")
            return

        item = _getUniqueCurveOrHistogram(plot)
        if item is None:
            # ambiguous case, we need to ask which plot item to fit
            isd = ItemsSelectionDialog(parent=plot, plot=plot)
            isd.setWindowTitle("Select item to be fitted")
            isd.setItemsSelectionMode(qt.QTableWidget.SingleSelection)
            isd.setAvailableKinds(["curve", "histogram"])
            isd.selectAllKinds()

            if not isd.exec():  # Cancel
                self._getToolWindow().setVisible(False)
            else:
                selectedItems = isd.getSelectedItems()
                item = selectedItems[0] if len(selectedItems) == 1 else None

        self._setXRange(*plot.getXAxis().getLimits())
        self._setFittedItem(item)

    def __updateFitWidget(self):
        """Update the data/range used by the FitWidget"""
        fitWidget = self._getToolWindow()

        item = self._getFittedItem()
        xdata = self.getXData(copy=False)
        ydata = self.getYData(copy=False)
        if item is None or xdata is None or ydata is None:
            fitWidget.setData(y=None)
            fitWidget.setWindowTitle("No curve selected")

        else:
            xmin, xmax = self.getXRange()
            fitWidget.setData(xdata, ydata, xmin=xmin, xmax=xmax)
            fitWidget.setWindowTitle(
                "Fitting " + item.getName() + " on x range %f-%f" % (xmin, xmax)
            )

    # X Range management

    def getXRange(self):
        """Returns the range on the X axis on which to perform the fit."""
        return self.__range

    def _setXRange(self, xmin, xmax):
        """Set the range on which the fit is done.

        :param float xmin:
        :param float xmax:
        """
        range_ = float(xmin), float(xmax)
        if self.__range != range_:
            self.__range = range_
            self.__updateFitWidget()

    def __setAutoXRangeEnabled(self, enabled):
        """Implement the change of update mode of the X range.

        :param bool enabled:
        """
        plot = self.plot
        if plot is None:
            _logger.error("No associated PlotWidget")
            return

        if enabled:
            self._setXRange(*plot.getXAxis().getLimits())
            plot.getXAxis().sigLimitsChanged.connect(self._setXRange)
        else:
            plot.getXAxis().sigLimitsChanged.disconnect(self._setXRange)

    def setXRangeUpdatedOnZoom(self, enabled):
        """Set whether or not to update the X range on zoom change.

        :param bool enabled:
        """
        if enabled != self.__rangeAutoUpdate:
            self.__rangeAutoUpdate = enabled
            if self._getToolWindow().isVisible():
                self.__setAutoXRangeEnabled(enabled)

    def isXRangeUpdatedOnZoom(self):
        """Returns the current mode of fitted data X range update.

        :rtype: bool
        """
        return self.__rangeAutoUpdate

    # Fitted item update

    def getXData(self, copy=True):
        """Returns the X data used for the fit or None if undefined.

        :param bool copy:
            True to get a copy of the data, False to get the internal data.
        :rtype: Union[numpy.ndarray,None]
        """
        return None if self.__x is None else numpy.array(self.__x, copy=copy or NP_OPTIONAL_COPY)

    def getYData(self, copy=True):
        """Returns the Y data used for the fit or None if undefined.

        :param bool copy:
            True to get a copy of the data, False to get the internal data.
        :rtype: Union[numpy.ndarray,None]
        """
        return None if self.__y is None else numpy.array(self.__y, copy=copy or NP_OPTIONAL_COPY)

    def _getFittedItem(self):
        """Returns the current item used for the fit

        :rtype: Union[~silx.gui.plot.items.Curve,~silx.gui.plot.items.Histogram,None]
        """
        return self.__item

    def _setFittedItem(self, item):
        """Set the curve to use for fitting.

        :param Union[~silx.gui.plot.items.Curve,~silx.gui.plot.items.Histogram,None] item:
        """
        plot = self.plot
        if plot is None:
            _logger.error("No associated PlotWidget")

        if plot is None or item is None:
            self.__item = None
            self.__curveParams = {}
            self.__updateFitWidget()
            return

        axis = item.getYAxis() if isinstance(item, items.YAxisMixIn) else "left"
        self.__curveParams = {
            "yaxis": axis,
            "xlabel": plot.getXAxis().getLabel(),
            "ylabel": plot.getYAxis(axis).getLabel(),
        }
        self.__legend = item.getName()

        if isinstance(item, items.Histogram):
            bin_edges = item.getBinEdgesData(copy=False)
            # take the middle coordinate between adjacent bin edges
            self.__x = (bin_edges[1:] + bin_edges[:-1]) / 2
            self.__y = item.getValueData(copy=False)
        # else take the active curve, or else the unique curve
        elif isinstance(item, items.Curve):
            self.__x = item.getXData(copy=False)
            self.__y = item.getYData(copy=False)

        self.__item = item
        self.__updateFitWidget()

    def __setFittedItemAutoUpdateEnabled(self, enabled):
        """Implement the change of fitted item update mode

        :param bool enabled:
        """
        plot = self.plot
        if plot is None:
            _logger.error("No associated PlotWidget")
            return

        self.__fitItemSelector.setPlotWidget(self.plot if enabled else None)

    def setFittedItemUpdatedFromActiveCurve(self, enabled):
        """Toggle fitted data synchronization with plot active curve.

        :param bool enabled:
        """
        enabled = bool(enabled)
        if enabled != self.__activeCurveSynchroEnabled:
            self.__activeCurveSynchroEnabled = enabled
            if self._getToolWindow().isVisible():
                self.__setFittedItemAutoUpdateEnabled(enabled)

    def isFittedItemUpdatedFromActiveCurve(self):
        """Returns True if fitted data is synchronized with plot.

        :rtype: bool
        """
        return self.__activeCurveSynchroEnabled

    # Handle fit completed

    def handle_signal(self, ddict):
        xdata = self.getXData(copy=False)
        if xdata is None:
            _logger.error("No reference data to display fit result for")
            return

        xmin, xmax = self.getXRange()
        x_fit = xdata[xmin <= xdata]
        x_fit = x_fit[x_fit <= xmax]
        fit_legend = "Fit <%s>" % self.__legend
        fit_curve = self.plot.getCurve(fit_legend)

        if ddict["event"] == "FitFinished":
            fit_widget = self._getToolWindow()
            if fit_widget is None:
                return
            y_fit = fit_widget.fitmanager.gendata()
            if fit_curve is None:
                self.plot.addCurve(
                    x_fit, y_fit, fit_legend, resetzoom=False, **self.__curveParams
                )
            else:
                fit_curve.setData(x_fit, y_fit)
                fit_curve.setVisible(True)
                fit_curve.setYAxis(self.__curveParams.get("yaxis", "left"))

        if ddict["event"] in ["FitStarted", "FitFailed"]:
            if fit_curve is not None:
                fit_curve.setVisible(False)



# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Basic tests for MaskToolsWidget"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "17/01/2018"


import logging
import os.path

import numpy

from silx.gui import qt
from silx.test.utils import temp_dir
from silx.utils.testutils import ParametricTestCase
from silx.gui.utils.testutils import getQToolButtonFromAction
from silx.gui.plot import PlotWindow, MaskToolsWidget
from .utils import PlotWidgetTestCase


_logger = logging.getLogger(__name__)


class TestMaskToolsWidget(PlotWidgetTestCase, ParametricTestCase):
    """Basic test for MaskToolsWidget"""

    def _createPlot(self):
        return PlotWindow()

    def setUp(self):
        super(TestMaskToolsWidget, self).setUp()
        self.widget = MaskToolsWidget.MaskToolsDockWidget(plot=self.plot, name="TEST")
        self.plot.addDockWidget(qt.Qt.BottomDockWidgetArea, self.widget)
        self.maskWidget = self.widget.widget()

    def tearDown(self):
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        del self.maskWidget
        del self.widget
        super(TestMaskToolsWidget, self).tearDown()

    def testEmptyPlot(self):
        """Empty plot, display MaskToolsDockWidget, toggle multiple masks"""
        self.maskWidget.setMultipleMasks("single")
        self.qapp.processEvents()

        self.maskWidget.setMultipleMasks("exclusive")
        self.qapp.processEvents()

    def _drag(self):
        """Drag from plot center to offset position"""
        plot = self.plot.getWidgetHandle()
        xCenter, yCenter = plot.width() // 2, plot.height() // 2
        offset = min(plot.width(), plot.height()) // 10

        pos0 = xCenter, yCenter
        pos1 = xCenter + offset, yCenter + offset

        self.mouseMove(plot, pos=(0, 0))
        self.mouseMove(plot, pos=pos0)
        self.qapp.processEvents()
        self.mousePress(plot, qt.Qt.LeftButton, pos=pos0)
        self.qapp.processEvents()
        self.mouseMove(plot, pos=(pos0[0] + offset // 2, pos0[1] + offset // 2))
        self.mouseMove(plot, pos=pos1)
        self.qapp.processEvents()
        self.mouseRelease(plot, qt.Qt.LeftButton, pos=pos1)
        self.qapp.processEvents()
        self.mouseMove(plot, pos=(0, 0))

    def _drawPolygon(self):
        """Draw a star polygon in the plot"""
        plot = self.plot.getWidgetHandle()
        x, y = plot.width() // 2, plot.height() // 2
        offset = min(plot.width(), plot.height()) // 10

        star = [
            (x, y + offset),
            (x - offset, y - offset),
            (x + offset, y),
            (x - offset, y),
            (x + offset, y - offset),
            (x, y + offset),
        ]  # Close polygon

        self.mouseMove(plot, pos=(0, 0))
        for pos in star:
            self.mouseMove(plot, pos=pos)
            self.qapp.processEvents()
            self.mousePress(plot, qt.Qt.LeftButton, pos=pos)
            self.qapp.processEvents()
            self.mouseRelease(plot, qt.Qt.LeftButton, pos=pos)
            self.qapp.processEvents()

    def _drawPencil(self):
        """Draw a star polygon in the plot"""
        plot = self.plot.getWidgetHandle()
        x, y = plot.width() // 2, plot.height() // 2
        offset = min(plot.width(), plot.height()) // 10

        star = [
            (x, y + offset),
            (x - offset, y - offset),
            (x + offset, y),
            (x - offset, y),
            (x + offset, y - offset),
        ]

        self.mouseMove(plot, pos=(0, 0))
        for start, end in zip(star[:-1], star[1:]):
            self.mouseMove(plot, pos=start)
            self.mousePress(plot, qt.Qt.LeftButton, pos=start)
            self.qapp.processEvents()
            self.mouseMove(plot, pos=end)
            self.qapp.processEvents()
            self.mouseRelease(plot, qt.Qt.LeftButton, pos=end)
            self.qapp.processEvents()

    def _isMaskItemSync(self):
        """Check if masks from item and tools are sync or not"""
        if self.maskWidget.isItemMaskUpdated():
            return numpy.all(
                numpy.equal(
                    self.maskWidget.getSelectionMask(),
                    self.plot.getActiveImage().getMaskData(copy=False),
                )
            )
        else:
            return True

    def testWithAnImage(self):
        """Plot with an image: test MaskToolsWidget interactions"""

        # Add and remove a image (this should enable/disable GUI + change mask)
        self.plot.addImage(
            numpy.random.random(1024**2).reshape(1024, 1024), legend="test"
        )
        self.qapp.processEvents()

        self.plot.remove("test", kind="image")
        self.qapp.processEvents()

        tests = [
            ((0, 0), (1, 1)),
            ((1000, 1000), (1, 1)),
            ((0, 0), (-1, -1)),
            ((1000, 1000), (-1, -1)),
        ]

        for itemMaskUpdated in (False, True):
            for origin, scale in tests:
                with self.subTest(origin=origin, scale=scale):
                    self.maskWidget.setItemMaskUpdated(itemMaskUpdated)
                    self.plot.addImage(
                        numpy.arange(1024**2).reshape(1024, 1024),
                        legend="test",
                        origin=origin,
                        scale=scale,
                    )
                    self.qapp.processEvents()

                    self.assertEqual(
                        self.maskWidget.isItemMaskUpdated(), itemMaskUpdated
                    )

                    # Test draw rectangle #
                    toolButton = getQToolButtonFromAction(self.maskWidget.rectAction)
                    self.assertIsNot(toolButton, None)
                    self.mouseClick(toolButton, qt.Qt.LeftButton)

                    # mask
                    self.maskWidget.maskStateGroup.button(1).click()
                    self.qapp.processEvents()
                    self._drag()
                    self.assertFalse(
                        numpy.all(numpy.equal(self.maskWidget.getSelectionMask(), 0))
                    )
                    self.assertTrue(self._isMaskItemSync())

                    # unmask same region
                    self.maskWidget.maskStateGroup.button(0).click()
                    self.qapp.processEvents()
                    self._drag()
                    self.assertTrue(
                        numpy.all(numpy.equal(self.maskWidget.getSelectionMask(), 0))
                    )
                    self.assertTrue(self._isMaskItemSync())

                    # Test draw polygon #
                    toolButton = getQToolButtonFromAction(self.maskWidget.polygonAction)
                    self.assertIsNot(toolButton, None)
                    self.mouseClick(toolButton, qt.Qt.LeftButton)

                    # mask
                    self.maskWidget.maskStateGroup.button(1).click()
                    self.qapp.processEvents()
                    self._drawPolygon()
                    self.assertFalse(
                        numpy.all(numpy.equal(self.maskWidget.getSelectionMask(), 0))
                    )
                    self.assertTrue(self._isMaskItemSync())

                    # unmask same region
                    self.maskWidget.maskStateGroup.button(0).click()
                    self.qapp.processEvents()
                    self._drawPolygon()
                    self.assertTrue(
                        numpy.all(numpy.equal(self.maskWidget.getSelectionMask(), 0))
                    )
                    self.assertTrue(self._isMaskItemSync())

                    # Test draw pencil #
                    toolButton = getQToolButtonFromAction(self.maskWidget.pencilAction)
                    self.assertIsNot(toolButton, None)
                    self.mouseClick(toolButton, qt.Qt.LeftButton)

                    self.maskWidget.pencilSpinBox.setValue(30)
                    self.qapp.processEvents()

                    # mask
                    self.maskWidget.maskStateGroup.button(1).click()
                    self.qapp.processEvents()
                    self._drawPencil()
                    self.assertFalse(
                        numpy.all(numpy.equal(self.maskWidget.getSelectionMask(), 0))
                    )
                    self.assertTrue(self._isMaskItemSync())

                    # unmask same region
                    self.maskWidget.maskStateGroup.button(0).click()
                    self.qapp.processEvents()
                    self._drawPencil()
                    self.assertTrue(
                        numpy.all(numpy.equal(self.maskWidget.getSelectionMask(), 0))
                    )
                    self.assertTrue(self._isMaskItemSync())

                    # Test no draw tool #
                    toolButton = getQToolButtonFromAction(self.maskWidget.browseAction)
                    self.assertIsNot(toolButton, None)
                    self.mouseClick(toolButton, qt.Qt.LeftButton)

                    self.plot.clear()

    def __loadSave(self, file_format):
        """Plot with an image: test MaskToolsWidget operations"""
        self.plot.addImage(numpy.arange(1024**2).reshape(1024, 1024), legend="test")
        self.qapp.processEvents()

        # Draw a polygon mask
        toolButton = getQToolButtonFromAction(self.maskWidget.polygonAction)
        self.assertIsNot(toolButton, None)
        self.mouseClick(toolButton, qt.Qt.LeftButton)
        self._drawPolygon()

        ref_mask = self.maskWidget.getSelectionMask()
        self.assertFalse(numpy.all(numpy.equal(ref_mask, 0)))

        with temp_dir() as tmp:
            mask_filename = os.path.join(tmp, "mask." + file_format)
            self.maskWidget.save(mask_filename, file_format)

            self.maskWidget.resetSelectionMask()
            self.assertTrue(
                numpy.all(numpy.equal(self.maskWidget.getSelectionMask(), 0))
            )

            self.maskWidget.load(mask_filename)
            self.assertTrue(
                numpy.all(numpy.equal(self.maskWidget.getSelectionMask(), ref_mask))
            )

    def testLoadSaveNpy(self):
        self.__loadSave("npy")

    def testLoadSaveFit2D(self):
        self.__loadSave("msk")

    def testSigMaskChangedEmitted(self):
        self.plot.addImage(numpy.arange(512**2).reshape(512, 512), legend="test")
        self.plot.resetZoom()
        self.qapp.processEvents()

        l = []

        def slot():
            l.append(1)

        self.maskWidget.sigMaskChanged.connect(slot)

        # rectangle mask
        toolButton = getQToolButtonFromAction(self.maskWidget.rectAction)
        self.assertIsNot(toolButton, None)
        self.mouseClick(toolButton, qt.Qt.LeftButton)
        self.maskWidget.maskStateGroup.button(1).click()
        self.qapp.processEvents()
        self._drag()

        self.assertGreater(len(l), 0)



# /*##########################################################################
#
# Copyright (c) 2004-2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
:mod:`silx.gui.plot.actions.mode` provides a set of QAction relative to mouse
mode of a :class:`.PlotWidget`.

The following QAction are available:

- :class:`ZoomModeAction`
- :class:`PanModeAction`
"""

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "16/08/2017"


from silx.gui import qt

from ..tools.menus import ZoomEnabledAxesMenu
from . import PlotAction


class ZoomModeAction(PlotAction):
    """QAction controlling the zoom mode of a :class:`.PlotWidget`.

    :param plot: :class:`.PlotWidget` instance on which to operate
    :param parent: See :class:`QAction`
    """

    def __init__(self, plot, parent=None):
        super(ZoomModeAction, self).__init__(
            plot,
            icon="zoom",
            text="Zoom mode",
            tooltip="Zoom-in on mouse selection",
            triggered=self._actionTriggered,
            checkable=True,
            parent=parent,
        )

        self.__menu = ZoomEnabledAxesMenu(self.plot, self.plot)

        # Listen to interaction configuration change
        self.plot.interaction().sigChanged.connect(self._interactionChanged)
        # Init the state
        self._interactionChanged()

    def isAxesMenuEnabled(self) -> bool:
        """Returns whether the axes selection menu is enabled or not (default: False)"""
        return self.menu() is self.__menu

    def setAxesMenuEnabled(self, enabled: bool):
        """Toggle the availability of the axes selection menu (default: False)"""
        if enabled == self.isAxesMenuEnabled():
            return

        self.setMenu(self.__menu if enabled else None)

        # Update associated QToolButton's popupMode if any, this is not done at least with Qt5
        parent = self.parent()
        if not isinstance(parent, qt.QToolBar):
            return
        widget = parent.widgetForAction(self)
        if not isinstance(widget, qt.QToolButton):
            return
        widget.setPopupMode(
            qt.QToolButton.MenuButtonPopup if enabled else qt.QToolButton.DelayedPopup
        )
        widget.update()

    def _interactionChanged(self):
        plot = self.plot
        if plot is None:
            return

        self.setChecked(plot.getInteractiveMode()["mode"] == "zoom")

    def _actionTriggered(self, checked=False):
        plot = self.plot
        if plot is None:
            return

        plot.setInteractiveMode("zoom", source=self)


class PanModeAction(PlotAction):
    """QAction controlling the pan mode of a :class:`.PlotWidget`.

    :param plot: :class:`.PlotWidget` instance on which to operate
    :param parent: See :class:`QAction`
    """

    def __init__(self, plot, parent=None):
        super(PanModeAction, self).__init__(
            plot,
            icon="pan",
            text="Pan mode",
            tooltip="Pan the view",
            triggered=self._actionTriggered,
            checkable=True,
            parent=parent,
        )
        # Listen to mode change
        self.plot.sigInteractiveModeChanged.connect(self._modeChanged)
        # Init the state
        self._modeChanged(None)

    def _modeChanged(self, source):
        modeDict = self.plot.getInteractiveMode()
        old = self.blockSignals(True)
        self.setChecked(modeDict["mode"] == "pan")
        self.blockSignals(old)

    def _actionTriggered(self, checked=False):
        plot = self.plot
        if plot is not None:
            plot.setInteractiveMode("pan", source=self)



# /*##########################################################################
#
# Copyright (c) 2004-2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Utility functions, toolbars and actions  to create profile on images
and stacks of images"""


__authors__ = ["V.A. Sole", "T. Vincent", "P. Knobel", "H. Payno"]
__license__ = "MIT"
__date__ = "12/04/2019"


import weakref

from .. import qt
from . import actions
from .tools.profile import manager
from .tools.profile import rois
from silx.gui.widgets.MultiModeAction import MultiModeAction

from .tools import roi as roi_mdl
from silx.gui.plot import items


class _CustomProfileManager(manager.ProfileManager):
    """This custom profile manager uses a single predefined profile window
    if it is specified. Else the behavior is the same as the default
    ProfileManager"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__profileWindow = None
        self.__specializedProfileWindows = {}

    def setSpecializedProfileWindow(self, roiClass, profileWindow):
        """Set a profile window for a given class or ROI.

        Setting profileWindow to None removes the roiClass from the list.

        :param roiClass:
        :param profileWindow:
        """
        if profileWindow is None:
            self.__specializedProfileWindows.pop(roiClass, None)
        else:
            self.__specializedProfileWindows[roiClass] = profileWindow

    def setProfileWindow(self, profileWindow):
        self.__profileWindow = profileWindow

    def createProfileWindow(self, plot, roi):
        for (
            roiClass,
            specializedProfileWindow,
        ) in self.__specializedProfileWindows.items():
            if isinstance(roi, roiClass):
                return specializedProfileWindow

        if self.__profileWindow is not None:
            return self.__profileWindow
        else:
            return super(_CustomProfileManager, self).createProfileWindow(plot, roi)

    def clearProfileWindow(self, profileWindow):
        for specializedProfileWindow in self.__specializedProfileWindows.values():
            if profileWindow is specializedProfileWindow:
                profileWindow.setProfile(None)
                return

        if self.__profileWindow is not None:
            self.__profileWindow.setProfile(None)
        else:
            return super(_CustomProfileManager, self).clearProfileWindow(profileWindow)


class ProfileToolBar(qt.QToolBar):
    """QToolBar providing profile tools operating on a :class:`PlotWindow`.

    Attributes:

    - plot: Associated :class:`PlotWindow` on which the profile line is drawn.
    - actionGroup: :class:`QActionGroup` of available actions.

    To run the following sample code, a QApplication must be initialized.
    First, create a PlotWindow and add a :class:`ProfileToolBar`.

    >>> from silx.gui.plot import PlotWindow
    >>> from silx.gui.plot.Profile import ProfileToolBar

    >>> plot = PlotWindow()  # Create a PlotWindow
    >>> toolBar = ProfileToolBar(plot=plot)  # Create a profile toolbar
    >>> plot.addToolBar(toolBar)  # Add it to plot
    >>> plot.show()  # To display the PlotWindow with the profile toolbar

    :param plot: :class:`PlotWindow` instance on which to operate.
    :param profileWindow: Plot widget instance where to
                          display the profile curve or None to create one.
    :param parent: See :class:`QToolBar`.
    """

    def __init__(self, parent=None, plot=None, profileWindow=None):
        super(ProfileToolBar, self).__init__(parent)
        assert plot is not None

        self._plotRef = weakref.ref(plot)

        # If a profileWindow is defined,
        # It will be used to display all the profiles
        self._manager = self.createProfileManager(self, plot)
        self._manager.setProfileWindow(profileWindow)
        self._manager.setDefaultColorFromCursorColor(True)
        self._manager.setItemType(image=True)
        self._manager.setActiveItemTracking(True)

        # Actions
        self._browseAction = actions.mode.ZoomModeAction(plot, parent=self)
        self._browseAction.setVisible(False)
        self.freeLineAction = None
        self._createProfileActions()
        self._editor = self._manager.createEditorAction(self)

        # ActionGroup
        self.actionGroup = qt.QActionGroup(self)
        self.actionGroup.addAction(self._browseAction)
        self.actionGroup.addAction(self.hLineAction)
        self.actionGroup.addAction(self.vLineAction)
        self.actionGroup.addAction(self.lineAction)
        self.actionGroup.addAction(self._editor)

        modes = MultiModeAction(self)
        modes.addAction(self.hLineAction)
        modes.addAction(self.vLineAction)
        modes.addAction(self.lineAction)
        if self.freeLineAction is not None:
            modes.addAction(self.freeLineAction)
        modes.addAction(self.crossAction)
        self.__multiAction = modes

        # Add actions to ToolBar
        self.addAction(self._browseAction)
        self.addAction(modes)
        self.addAction(self._editor)
        self.addAction(self.clearAction)

        plot.sigActiveImageChanged.connect(self._activeImageChanged)
        self._activeImageChanged()

    def createProfileManager(self, parent, plot):
        return _CustomProfileManager(parent, plot)

    def _createProfileActions(self):
        self.hLineAction = self._manager.createProfileAction(
            rois.ProfileImageHorizontalLineROI, self
        )
        self.vLineAction = self._manager.createProfileAction(
            rois.ProfileImageVerticalLineROI, self
        )
        self.lineAction = self._manager.createProfileAction(
            rois.ProfileImageLineROI, self
        )
        self.freeLineAction = self._manager.createProfileAction(
            rois.ProfileImageDirectedLineROI, self
        )
        self.crossAction = self._manager.createProfileAction(
            rois.ProfileImageCrossROI, self
        )
        self.clearAction = self._manager.createClearAction(self)

    def getPlotWidget(self):
        """The :class:`.PlotWidget` associated to the toolbar."""
        return self._plotRef()

    def _setRoiActionEnabled(self, itemKind, enabled):
        for action in self.__multiAction.getMenu().actions():
            if not isinstance(action, roi_mdl.CreateRoiModeAction):
                continue
            roiClass = action.getRoiClass()
            if issubclass(itemKind, roiClass.ITEM_KIND):
                action.setEnabled(enabled)

    def _activeImageChanged(self, previous=None, legend=None):
        """Handle active image change to toggle actions"""
        if legend is None:
            self._setRoiActionEnabled(items.ImageStack, False)
            self._setRoiActionEnabled(items.ImageBase, False)
        else:
            plot = self.getPlotWidget()
            image = plot.getActiveImage()
            # Disable for empty image
            enabled = image.getData(copy=False).size > 0
            self._setRoiActionEnabled(type(image), enabled)

    def getProfileManager(self):
        """Return the manager of the profiles.

        :rtype: ProfileManager
        """
        return self._manager

    def clearProfile(self):
        """Remove profile curve and profile area."""
        self._manager.clearProfile()


class Profile3DToolBar(ProfileToolBar):
    def __init__(self, parent=None, stackview=None):
        """QToolBar providing profile tools for an image or a stack of images.

        :param parent: the parent QWidget
        :param stackview: :class:`StackView` instance on which to operate.
        :param parent: See :class:`QToolBar`.
        """
        # TODO: add param profileWindow (specify the plot used for profiles)
        super(Profile3DToolBar, self).__init__(
            parent=parent, plot=stackview.getPlotWidget()
        )

        self.stackView = stackview
        """:class:`StackView` instance"""

    def _createProfileActions(self):
        self.hLineAction = self._manager.createProfileAction(
            rois.ProfileImageStackHorizontalLineROI, self
        )
        self.vLineAction = self._manager.createProfileAction(
            rois.ProfileImageStackVerticalLineROI, self
        )
        self.lineAction = self._manager.createProfileAction(
            rois.ProfileImageStackLineROI, self
        )
        self.crossAction = self._manager.createProfileAction(
            rois.ProfileImageStackCrossROI, self
        )
        self.clearAction = self._manager.createClearAction(self)



#!/usr/bin/env python3
# /*##########################################################################
#
# Copyright (c) 2016-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Script generating an HDF5 containing most of special structures supported by
the format.
"""

__license__ = "MIT"
__date__ = "26/03/2019"


import json
import logging
import sys

logging.basicConfig()
logger = logging.getLogger("create_h5file")


try:
    import hdf5plugin
except ImportError:
    logger.error("Backtrace", exc_info=True)

import h5py
import numpy


if sys.version_info.major < 3:
    raise RuntimeError("Python 2 is not supported")


def str_to_utf8(text):
    """Convert str or sequence of str to type compatible with h5py

    :param Union[str,List[str]] text:
    :rtype: numpy.ndarray
    """
    return numpy.array(text, dtype=h5py.special_dtype(vlen=str))


def store_subdimensions(group, data, dtype, prefix=None):
    """Creates datasets in given group with data

    :param h5py.Group group: Group where to add the datasets
    :param numpy.ndarray data: The data to use
    :param Union[numpy.dtype,str] dtype:
    :param Union[str,None] prefix: String to use as datasets name prefix
    """
    try:
        dtype = numpy.dtype(dtype)
    except TypeError as e:
        logger.error("Cannot create datasets for dtype: %s", str(dtype))
        logger.error(e)
        return

    if prefix is None:
        prefix = str(dtype)

    if hasattr(h5py, "Empty"):
        basename = prefix + "_empty"
        try:
            group[basename] = h5py.Empty(dtype=numpy.dtype(dtype))
        except (RuntimeError, ValueError) as e:
            logger.error("Error while creating %s in %s" % (basename, str(group)))
            logger.error(e)
    else:
        logger.warning("h5py.Empty not available")

    data = data.astype(dtype)
    data.shape = -1
    basename = prefix + "_d0"
    try:
        group[basename] = data[0]
    except RuntimeError as e:
        logger.error("Error while creating %s in %s" % (basename, str(group)))
        logger.error(e)

    shapes = [10, 4, 4, 4]
    for i in range(1, 4):
        shape = shapes[:i]
        shape.append(-1)
        reversed(shape)
        shape = tuple(shape)
        data.shape = shape
        basename = prefix + "_d%d" % i
        try:
            group[basename] = data
        except RuntimeError as e:
            logger.error("Error while creating %s in %s" % (basename, str(group)))
            logger.error(e)


def create_hdf5_types(group):
    print("- Creating HDF types...")

    main_group = group.create_group("HDF5")

    # H5T_INTEGER

    int_data = numpy.random.randint(-100, 100, size=10 * 4 * 4 * 4)
    uint_data = numpy.random.randint(0, 100, size=10 * 4 * 4 * 4)

    group = main_group.create_group("integer_little_endian")
    for size in (1, 2, 4, 8):
        store_subdimensions(
            group, int_data, "<i" + str(size), prefix="int" + str(size * 8)
        )
        store_subdimensions(
            group, uint_data, "<u" + str(size), prefix="uint" + str(size * 8)
        )

    group = main_group.create_group("integer_big_endian")
    for size in (1, 2, 4, 8):
        store_subdimensions(
            group, int_data, ">i" + str(size), prefix="int" + str(size * 8)
        )
        store_subdimensions(
            group, uint_data, ">u" + str(size), prefix="uint" + str(size * 8)
        )

    # H5T_FLOAT

    float_data = numpy.random.rand(10 * 4 * 4 * 4)
    group = main_group.create_group("float_little_endian")

    for size in (2, 4, 8):
        store_subdimensions(
            group, float_data, "<f" + str(size), prefix="float" + str(size * 8)
        )

    group = main_group.create_group("float_big_endian")

    for size in (2, 4, 8):
        store_subdimensions(
            group, float_data, ">f" + str(size), prefix="float" + str(size * 8)
        )

    # H5T_TIME

    main_group.create_group("time")

    # H5T_STRING

    group = main_group.create_group("text")
    group.create_dataset("ascii_vlen", data=b"abcd", dtype=h5py.string_dtype("ascii"))
    group.create_dataset(
        "bad_ascii_vlen", data=b"ab\xEFcd\xFF", dtype=h5py.string_dtype("ascii")
    )
    group.create_dataset(
        "ascii_fixed", data=b"Fixed ascii", dtype=h5py.string_dtype("ascii", 20)
    )
    group.create_dataset(
        "array_ascii_vlen",
        data=["a", "bc", "def", "ghij", "klmn"],
        dtype=h5py.string_dtype("ascii"),
    )
    group.create_dataset(
        "array_ascii_fixed",
        data=["abcde", "fghij"],
        dtype=h5py.string_dtype("ascii", 5),
    )

    group.create_dataset(
        "utf8_vlen", data="me \u2661 tu", dtype=h5py.string_dtype("utf-8")
    )
    group.create_dataset(
        "utf8_fixed",
        data="me \u2661 tu".encode("utf-8"),
        dtype=h5py.string_dtype("utf-8", 20),
    )
    group.create_dataset(
        "array_utf8_vlen",
        data=["a", "bc", "def", "ghij", "klmn"],
        dtype=h5py.string_dtype("utf-8"),
    )
    group.create_dataset(
        "array_utf8_fixed", data=["abcde", "fghij"], dtype=h5py.string_dtype("utf-8", 5)
    )

    # H5T_BITFIELD

    main_group.create_group("bitfield")

    # H5T_OPAQUE

    group = main_group.create_group("opaque")

    main_group["opaque/ascii"] = numpy.void(b"abcd")
    main_group["opaque/utf8"] = numpy.void("i \u2661 my mother".encode("utf-8"))
    main_group["opaque/thing"] = numpy.void(b"\x10\x20\x30\x40\xF0")
    main_group["opaque/big_thing"] = numpy.void(b"\x10\x20\x30\x40\xF0" * 100000)

    data = numpy.void(b"\x10\x20\x30\x40\xFF" * 20)
    data = numpy.array([data] * 10 * 4 * 4 * 4, numpy.void)
    store_subdimensions(group, data, "void")

    # H5T_COMPOUND

    a = numpy.array(
        [(1, 2.0, "Hello"), (2, 3.0, "World")],
        dtype=[("foo", "i4"), ("bar", "f4"), ("baz", "S10")],
    )

    b = numpy.zeros(3, dtype="3int8, float32, (2,3)float64")

    c = numpy.zeros(
        3, dtype=("i4", [("r", "u1"), ("g", "u1"), ("b", "u1"), ("a", "u1")])
    )

    d = numpy.zeros(
        3, dtype=[("x", "f4"), ("y", numpy.float32), ("value", "f4", (2, 2))]
    )

    e = numpy.zeros(3, dtype={"names": ["col1", "col2"], "formats": ["i4", "f4"]})

    f = numpy.array(
        [(1.5, 2.5, (1.0, 2.0)), (3.0, 4.0, (4.0, 5.0)), (1.0, 3.0, (2.0, 6.0))],
        dtype=[("x", "f4"), ("y", numpy.float32), ("value", "f4", (2, 2))],
    )

    main_group["compound/numpy_example_a"] = a
    main_group["compound/numpy_example_b"] = b
    main_group["compound/numpy_example_c"] = c
    main_group["compound/numpy_example_d"] = d
    main_group["compound/numpy_example_e"] = e
    main_group["compound/numpy_example_f"] = f

    dt = numpy.dtype([("start", numpy.uint32), ("stop", numpy.uint32)])
    vlen_dt = h5py.special_dtype(vlen=dt)
    data = numpy.array([[(1, 2), (2, 3)], [(3, 5), (5, 8), (8, 9)]], vlen_dt)
    dataset = main_group.create_dataset("compound/vlen", data.shape, data.dtype)
    for i, row in enumerate(data):
        dataset[i] = row

    # numpy complex is a H5T_COMPOUND

    real_data = numpy.random.rand(10 * 4 * 4 * 4)
    imaginary_data = numpy.random.rand(10 * 4 * 4 * 4)
    complex_data = real_data + imaginary_data * 1j

    group = main_group.create_group("compound/numpy_complex_little_endian")
    for size in (8, 16, 32):
        store_subdimensions(
            group, complex_data, "<c" + str(size), prefix="complex" + str(size * 8)
        )

    group = main_group.create_group("compound/numpy_complex_big_endian")
    for size in (8, 16, 32):
        store_subdimensions(
            group, complex_data, ">c" + str(size), prefix="complex" + str(size * 8)
        )

    # H5T_REFERENCE

    ref_dt = h5py.special_dtype(ref=h5py.Reference)
    group = main_group.create_group("reference")
    data = group.create_dataset("data", data=numpy.random.rand(10, 10))

    group.create_dataset("ref_0d", data=data.ref, dtype=ref_dt)
    group.create_dataset("ref_1d", data=[data.ref, None], dtype=ref_dt)
    group.create_dataset("regionref_0d", data=data.regionref[0:10, 0:5], dtype=ref_dt)
    group.create_dataset("regionref_1d", data=[data.regionref[0:10, 0:5]], dtype=ref_dt)

    # H5T_ENUM

    enum_dt = h5py.special_dtype(enum=("i", {"RED": 0, "GREEN": 1, "BLUE": 42}))
    group = main_group.create_group("enum")
    uint_data = numpy.random.randint(0, 100, size=10 * 4 * 4 * 4)
    uint_data.shape = 10, 4, 4, 4

    group.create_dataset("color_0d", data=numpy.array(42, dtype=enum_dt))
    group.create_dataset("color_1d", data=numpy.array([0, 1, 100, 42], dtype=enum_dt))
    group.create_dataset("color_4d", data=numpy.array(uint_data, dtype=enum_dt))

    # numpy bool is a H5T_ENUM

    bool_data = uint_data < 50
    bool_group = main_group.create_group("enum/numpy_boolean")
    store_subdimensions(bool_group, bool_data, "bool")

    # H5T_VLEN

    group = main_group.create_group("vlen")
    text = "i \u2661 my dad"

    unicode_vlen_dt = h5py.special_dtype(vlen=str)
    group.create_dataset("unicode", data=numpy.array(text, dtype=unicode_vlen_dt))
    group.create_dataset("unicode_1d", data=numpy.array([text], dtype=unicode_vlen_dt))


def set_silx_style(group, axes_scales=None, signal_scale=None):
    """Set SILX_style attribute

    :param group: NXdata group to set SILX_style for.
    :param axes_scales: Scale to use for the axes.
        Used for plot axis scales.
    :param signal_scale: Scale to use for the signal.
        Used either for plot axis scale or colormap normalization.
    """
    style = {}
    if axes_scales is not None:
        style["axes_scale_types"] = axes_scales
    if signal_scale is not None:
        style["signal_scale_type"] = signal_scale
    if style:
        group.attrs["SILX_style"] = json.dumps(style)


def create_nxdata_group(group):
    print("- Creating NXdata types...")

    main_group = group.create_group("NxData")

    # SCALARS
    g0d = main_group.create_group("scalars")

    g0d0 = g0d.create_group("0D_scalar")
    g0d0.attrs["NX_class"] = "NXdata"
    g0d0.attrs["signal"] = "scalar"
    g0d0.create_dataset("scalar", data=10)

    g0d1 = g0d.create_group("2D_scalars")
    g0d1.attrs["NX_class"] = "NXdata"
    g0d1.attrs["signal"] = "scalars"
    ds = g0d1.create_dataset("scalars", data=numpy.arange(3 * 10).reshape((3, 10)))
    ds.attrs["interpretation"] = "scalar"

    g0d1 = g0d.create_group("4D_scalars")
    g0d1.attrs["NX_class"] = "NXdata"
    g0d1.attrs["signal"] = "scalars"
    ds = g0d1.create_dataset(
        "scalars", data=numpy.arange(2 * 2 * 3 * 10).reshape((2, 2, 3, 10))
    )
    ds.attrs["interpretation"] = "scalar"

    # SPECTRA
    g1d = main_group.create_group("spectra")

    g1d0 = g1d.create_group("1D_spectrum")
    g1d0.attrs["NX_class"] = "NXdata"
    g1d0.attrs["signal"] = "count"
    g1d0.attrs["axes"] = "energy_calib"
    g1d0.attrs["uncertainties"] = str_to_utf8(("energy_errors",))
    g1d0.create_dataset("count", data=numpy.arange(10))
    g1d0.create_dataset("energy_calib", data=(10, 5))  # 10 * idx + 5
    g1d0.create_dataset("energy_errors", data=3.14 * numpy.random.rand(10))
    set_silx_style(g1d0, axes_scales=["linear"], signal_scale="log")

    g1d1 = g1d.create_group("2D_spectra")
    g1d1.attrs["NX_class"] = "NXdata"
    g1d1.attrs["signal"] = "counts"
    ds = g1d1.create_dataset("counts", data=numpy.arange(3 * 10).reshape((3, 10)))
    ds.attrs["interpretation"] = "spectrum"
    set_silx_style(g1d1, axes_scales=["linear"], signal_scale="log")

    g1d2 = g1d.create_group("4D_spectra")
    g1d2.attrs["NX_class"] = "NXdata"
    g1d2.attrs["signal"] = "counts"
    g1d2.attrs["axes"] = str_to_utf8(("energy",))
    ds = g1d2.create_dataset(
        "counts", data=numpy.arange(2 * 2 * 3 * 10).reshape((2, 2, 3, 10))
    )
    ds.attrs["interpretation"] = "spectrum"
    ds = g1d2.create_dataset("errors", data=4.5 * numpy.random.rand(2, 2, 3, 10))
    ds = g1d2.create_dataset(
        "energy", data=5 + 10 * numpy.arange(15), shuffle=True, compression="gzip"
    )
    ds.attrs["long_name"] = "Calibrated energy"
    ds.attrs["first_good"] = 3
    ds.attrs["last_good"] = 12
    g1d2.create_dataset("energy_errors", data=10 * numpy.random.rand(15))
    set_silx_style(g1d2, axes_scales=["linear"], signal_scale="log")

    # IMAGES
    g2d = main_group.create_group("images")

    g2d0 = g2d.create_group("2D_regular_image")
    g2d0.attrs["NX_class"] = "NXdata"
    g2d0.attrs["signal"] = "image"
    g2d0.attrs["axes"] = str_to_utf8(("rows_calib", "columns_coordinates"))
    g2d0.create_dataset("image", data=numpy.arange(4 * 6).reshape((4, 6)))
    ds = g2d0.create_dataset("rows_calib", data=(10, 5))
    ds.attrs["long_name"] = "Calibrated Y"
    g2d0.create_dataset("columns_coordinates", data=0.5 + 0.02 * numpy.arange(6))
    set_silx_style(g1d2, axes_scales=["linear"], signal_scale="log")

    g2d1 = g2d.create_group("2D_irregular_data")
    g2d1.attrs["NX_class"] = "NXdata"
    g2d1.attrs["signal"] = "data"
    g2d1.attrs["axes"] = str_to_utf8(("rows_coordinates", "columns_coordinates"))
    g2d1.create_dataset("data", data=numpy.arange(64 * 128).reshape((64, 128)))
    g2d1.create_dataset(
        "rows_coordinates", data=numpy.arange(64) + numpy.random.rand(64)
    )
    g2d1.create_dataset(
        "columns_coordinates", data=numpy.arange(128) + 2.5 * numpy.random.rand(128)
    )
    set_silx_style(g2d1, axes_scales=["linear", "log"], signal_scale="log")

    g2d2 = g2d.create_group("3D_images")
    g2d2.attrs["NX_class"] = "NXdata"
    g2d2.attrs["signal"] = "images"
    ds = g2d2.create_dataset("images", data=numpy.arange(2 * 4 * 6).reshape((2, 4, 6)))
    ds.attrs["interpretation"] = "image"
    set_silx_style(g2d2, signal_scale="log")

    g2d3 = g2d.create_group("5D_images")
    g2d3.attrs["NX_class"] = "NXdata"
    g2d3.attrs["signal"] = "images"
    g2d3.attrs["axes"] = str_to_utf8(("rows_coordinates", "columns_coordinates"))
    ds = g2d3.create_dataset(
        "images", data=numpy.arange(2 * 2 * 2 * 4 * 6).reshape((2, 2, 2, 4, 6))
    )
    ds.attrs["interpretation"] = "image"
    g2d3.create_dataset("rows_coordinates", data=5 + 10 * numpy.arange(4))
    g2d3.create_dataset("columns_coordinates", data=0.5 + 0.02 * numpy.arange(6))
    set_silx_style(g2d3, signal_scale="log")

    y = numpy.arange(-5, 10).reshape(-1, 1)
    x = numpy.arange(-5, 10).reshape(1, -1)
    data = x**2 + y**2
    data = data + (data.max() - data) * 1j

    g2d3 = g2d.create_group("2D_complex_image")
    g2d3.attrs["NX_class"] = "NXdata"
    g2d3.attrs["signal"] = "image"
    g2d3.attrs["axes"] = str_to_utf8(("rows", "columns"))
    g2d3.create_dataset("image", data=data)
    g2d3.create_dataset("rows", data=0.5 + 0.02 * numpy.arange(data.shape[0]))
    g2d3.create_dataset("columns", data=0.5 + 0.02 * numpy.arange(data.shape[1]))
    set_silx_style(g2d3, signal_scale="log")

    # SCATTER
    g = main_group.create_group("scatters")

    gd0 = g.create_group("x_y_scatter")
    gd0.attrs["NX_class"] = "NXdata"
    gd0.attrs["signal"] = "y"
    gd0.attrs["axes"] = str_to_utf8(("x",))
    gd0.create_dataset("y", data=numpy.random.rand(128) - 0.5)
    gd0.create_dataset("x", data=2 * numpy.random.rand(128))
    gd0.create_dataset("x_errors", data=0.05 * numpy.random.rand(128))
    gd0.create_dataset("errors", data=0.05 * numpy.random.rand(128))
    set_silx_style(gd0, axes_scales=["linear"], signal_scale="log")

    gd1 = g.create_group("x_y_value_scatter")
    gd1.attrs["NX_class"] = "NXdata"
    gd1.attrs["signal"] = "values"
    gd1.attrs["axes"] = str_to_utf8(("x", "y"))
    gd1.create_dataset("values", data=3.14 * numpy.random.rand(128))
    gd1.create_dataset("y", data=numpy.random.rand(128))
    gd1.create_dataset("y_errors", data=0.02 * numpy.random.rand(128))
    gd1.create_dataset("x", data=numpy.random.rand(128))
    gd1.create_dataset("x_errors", data=0.02 * numpy.random.rand(128))
    set_silx_style(gd1, axes_scales=["linear", "log"], signal_scale="log")

    # NDIM > 3
    g = main_group.create_group("cubes")

    gd0 = g.create_group("3D_cube")
    gd0.attrs["NX_class"] = "NXdata"
    gd0.attrs["signal"] = "cube"
    gd0.attrs["axes"] = str_to_utf8(("img_idx", "rows_coordinates", "cols_coordinates"))
    gd0.create_dataset("cube", data=numpy.arange(4 * 5 * 6).reshape((4, 5, 6)))
    gd0.create_dataset("img_idx", data=numpy.arange(4))
    gd0.create_dataset("rows_coordinates", data=0.1 * numpy.arange(5))
    gd0.create_dataset("cols_coordinates", data=[0.2, 0.3])  # linear calibration
    set_silx_style(gd0, axes_scales=["log", "linear"], signal_scale="log")

    gd1 = g.create_group("5D")
    gd1.attrs["NX_class"] = "NXdata"
    gd1.attrs["signal"] = "hypercube"
    data = numpy.arange(2 * 3 * 4 * 5 * 6).reshape((2, 3, 4, 5, 6))
    gd1.create_dataset("hypercube", data=data)
    set_silx_style(gd1, axes_scales=["log", "linear"], signal_scale="log")


FILTER_LZF = 32000
FILTER_LZ4 = 32004
FILTER_CBF = 32006
FILTER_BITSHUFFLE = 32008
FILTER_BITSHUFFLE_COMPRESS_LZ4 = 2
FILTER_USER = 32768
"""First id for non-distributed filter"""

encoded_data = [
    (
        "lzf",
        {"compression": FILTER_LZF},
        b"\x01\x00\x00\x80\x00\x00\x01\x80\x06\x01\x00\x02\xa0\x07\x00\x03\xa0\x07"
        b"\x00\x04\xa0\x07\x00\x05\xa0\x07\x00\x06\xa0\x07\x00\x07\xa0\x07\x00\x08"
        b"\xa0\x07\x00\t\xa0\x07\x00\n\xa0\x07\x00\x0b\xa0\x07\x00\x0c\xa0\x07\x00"
        b"\r\xa0\x07\x00\x0e\xa0\x07\x00\x0f`\x07\x01\x00\x00",
    ),
    (
        "bitshuffle+lz4",
        {
            "compression": FILTER_BITSHUFFLE,
            "compression_opts": (0, FILTER_BITSHUFFLE_COMPRESS_LZ4),
        },
        b"\x00\x00\x00\x00\x00\x00\x00\x80\x00\x00 \x00\x00\x00\x00\x13\x9f\xaa\xaa"
        b"\xcc\xcc\xf0\xf0\x00\xff\x00\x01\x00_P\x00\x00\x00\x00\x00",
    ),
    ("cbf", {"compression": FILTER_CBF}, None),
    (
        "lz4",
        {"compression": FILTER_LZ4},
        b"\x00\x00\x00\x00\x00\x00\x00\x80\x00\x00\x00\x80\x00\x00\x00E\x13\x00\x01"
        b"\x00\x13\x01\x08\x00\x13\x02\x08\x00\x13\x03\x08\x00\x13\x04\x08\x00\x13"
        b"\x05\x08\x00\x13\x06\x08\x00\x13\x07\x08\x00\x13\x08\x08\x00\x13\t\x08"
        b"\x00\x13\n\x08\x00\x13\x0b\x08\x00\x13\x0c\x08\x00\x13\r\x08\x00\x13\x0e"
        b"\x08\x00\x80\x0f\x00\x00\x00\x00\x00\x00\x00",
    ),
    (
        "corrupted_datasets/bitshuffle+lz4",
        {
            "compression": FILTER_BITSHUFFLE,
            "compression_opts": (0, FILTER_BITSHUFFLE_COMPRESS_LZ4),
        },
        b"\xFF\x01\x00\x01" * 10,
    ),
    (
        "corrupted_datasets/unavailable_filter",
        {"compression": FILTER_USER + 100},
        b"\xFF\x01\x00\x01" * 10,
    ),
]


def display_encoded_data():
    reference = numpy.arange(16, dtype=int).reshape(4, 4)

    import os
    import tempfile

    tempdir = tempfile.mkdtemp()
    filename = os.path.join(tempdir, "encode.h5")
    for info in encoded_data:
        name, compression_args, stored_data = info
        print("m" * 20)
        print(name)
        try:
            with h5py.File(filename, "a") as h5:
                dataset = h5.create_dataset(
                    name, data=reference, chunks=reference.shape, **compression_args
                )
                offsets = (0,) * len(reference.shape)
                filter_mask = [0xFFFF]
                data = dataset.id.read_direct_chunk(
                    offsets=offsets, filter_mask=filter_mask
                )
            if data != stored_data:
                print("- Data changed")
            print("- Data:")
            print(data)
        except Exception:
            logger.error("Backtrace", exc_info=True)


def create_encoded_data(parent_group):
    print("- Creating encoded data...")

    group = parent_group.create_group("encoded")

    reference = numpy.arange(16).reshape(4, 4)
    group["uncompressed"] = reference

    group.create_dataset("zipped", data=reference, compression="gzip")
    group.create_dataset("zipped_max", data=reference, compression="gzip")

    compressions = ["szip"]
    for compression in compressions:
        try:
            group.create_dataset(compression, data=reference, compression=compression)
        except Exception:
            logger.warning(
                "Filter '%s' is not available. Dataset skipped.", compression
            )

    group.create_dataset("shuffle_filter", data=reference, shuffle=True)
    group.create_dataset("fletcher32_filter", data=reference, fletcher32=True)

    for info in encoded_data:
        name, compression_args, data = info
        if data is None:
            logger.warning(
                "No compressed data for dataset '%s'. Dataset skipped.", name
            )
            continue
        try:
            dataset = group.create_dataset(
                name,
                shape=reference.shape,
                maxshape=reference.shape,
                dtype=reference.dtype,
                chunks=reference.shape,
                **compression_args,
            )
        except Exception:
            logger.debug("Backtrace", exc_info=True)
            logger.error("Error while creating dataset '%s'", name)
            continue
        try:
            offsets = (0,) * len(reference.shape)
            dataset.id.write_direct_chunk(
                offsets=offsets, data=data, filter_mask=0x0000
            )
        except Exception:
            logger.debug("Backtrace", exc_info=True)
            logger.error("Error filling dataset '%s'", name)
            continue


def create_links(h5):
    print("- Creating links...")

    main_group = h5.create_group("links")
    dataset = main_group.create_dataset("dataset", data=numpy.int64(10))
    group = main_group.create_group("group")
    group["something_inside"] = numpy.int64(20)

    main_group["hard_link_to_group"] = main_group["group"]
    main_group["hard_link_to_dataset"] = main_group["dataset"]

    main_group.create_group("hard_recursive_link")
    main_group.create_group("hard_recursive_link2")
    main_group["hard_recursive_link/link"] = main_group["hard_recursive_link2"]
    main_group["hard_recursive_link2/link"] = main_group["hard_recursive_link"]

    main_group["soft_link_to_group"] = h5py.SoftLink(group.name)
    main_group["soft_link_to_dataset"] = h5py.SoftLink(dataset.name)
    main_group["soft_link_to_nothing"] = h5py.SoftLink("/foo/bar/2000")
    main_group["soft_link_to_group_link"] = h5py.SoftLink(
        main_group.name + "/soft_link_to_group"
    )
    main_group["soft_link_to_dataset_link"] = h5py.SoftLink(
        main_group.name + "/soft_link_to_dataset"
    )
    main_group["soft_link_to_itself"] = h5py.SoftLink(
        main_group.name + "/soft_link_to_itself"
    )

    # External links to self file
    main_group["external_link_to_group"] = h5py.ExternalLink(
        h5.file.filename, group.name
    )
    main_group["external_link_to_dataset"] = h5py.ExternalLink(
        h5.file.filename, dataset.name
    )
    main_group["external_link_to_nothing"] = h5py.ExternalLink(
        h5.file.filename, "/foo/bar/2000"
    )
    main_group["external_link_to_missing_file"] = h5py.ExternalLink(
        h5.file.filename + "_unknown", "/"
    )
    main_group["external_link_to_group_link"] = h5py.ExternalLink(
        h5.file.filename, main_group.name + "/soft_link_to_group"
    )
    main_group["external_link_to_dataset_link"] = h5py.ExternalLink(
        h5.file.filename, main_group.name + "/soft_link_to_dataset"
    )
    main_group["external_link_to_itself"] = h5py.ExternalLink(
        h5.file.filename, main_group.name + "/external_link_to_itself"
    )
    main_group["external_link_to_recursive_link2"] = h5py.ExternalLink(
        h5.file.filename, main_group.name + "/external_link_to_recursive_link3"
    )
    main_group["external_link_to_recursive_link3"] = h5py.ExternalLink(
        h5.file.filename, main_group.name + "/external_link_to_recursive_link2"
    )
    main_group["external_link_to_soft_recursive"] = h5py.ExternalLink(
        h5.file.filename, main_group.name + "/soft_link_to_itself"
    )


def create_image_types(h5):
    print("- Creating images...")
    raw = numpy.arange(10 * 10).reshape((10, 10))
    u8 = (raw * 255 / raw.max()).astype(numpy.uint8)
    f32 = (raw / raw.max()).astype(numpy.float32)
    f64 = (raw / raw.max()).astype(numpy.float64)

    u8_t = numpy.swapaxes(u8, 0, 1)
    f32_t = numpy.swapaxes(f32, 0, 1)
    f64_t = numpy.swapaxes(f64, 0, 1)

    main_group = h5.create_group("images")

    main_group["intensity_uint8"] = u8
    main_group["intensity_float32"] = f32
    main_group["intensity_float64"] = f64
    main_group["rgb_uint8"] = numpy.stack((u8, u8_t, numpy.flip(u8, 0))).T
    main_group["rgb_float32"] = numpy.stack((f32, f32_t, numpy.flip(f32, 0))).T
    main_group["rgb_float64"] = numpy.stack((f64, f64_t, numpy.flip(f64, 0))).T
    main_group["rgba_uint8"] = numpy.stack(
        (u8, u8_t, numpy.flip(u8, 0), numpy.flip(u8_t))
    ).T
    main_group["rgba_float32"] = numpy.stack(
        (f32, f32_t, numpy.flip(f32, 0), numpy.flip(f32_t))
    ).T
    main_group["rgba_float64"] = numpy.stack(
        (f64, f64_t, numpy.flip(f64, 0), numpy.flip(f64_t))
    ).T


def create_file():
    filename = "all_types.h5"
    print("Creating file '%s'..." % filename)
    with h5py.File(filename, "w") as h5:
        create_hdf5_types(h5)
        create_nxdata_group(h5)
        create_encoded_data(h5)
        create_links(h5)
        create_image_types(h5)


def main():
    print("Begin")
    create_file()
    # display_encoded_data()
    print("End")


if __name__ == "__main__":
    main()
